CREATE OR REPLACE PACKAGE TabMan AS
-- TabMan: Table Manager V2.0
-- Author: Jorgen Bosman
-- e-mail: jbosman@village.uunet.be or jbosma@fercomsys.be

-- This package is designed to run on Oracle 8.0.5 or higher
-- (I think it should also run on Oracle 8.0.4, but untested)

-- Minimum required privileges (granted directly to the user, not via a role for some strange reason!!):
--  GRANT CREATE PUBLIC SYNONYM TO X;
--  GRANT CREATE TABLE TO X;
--  GRANT CREATE SYNONYM TO X;
--  GRANT CREATE TRIGGER TO X;
--  GRANT DROP PUBLIC SYNONYM TO X;
-- If other users must use this package then the follwing extra privileges are required:
--  GRANT ALTER ANY INDEX TO X;
--  GRANT ALTER ANY TABLE TO X;
--  GRANT ALTER ANY TRIGGER TO X;
--  GRANT COMMENT ANY TABLE TO X;
--  GRANT CREATE ANY INDEX TO X;
--  GRANT CREATE ANY TABLE TO X;
--  GRANT CREATE ANY TRIGGER TO X;
--  GRANT CREATE ANY SYNONYM TO X;
--  GRANT DROP ANY INDEX TO X;
--  GRANT DROP ANY SYNONYM TO X;
--  GRANT DROP ANY TABLE TO X;
--  GRANT DROP ANY TRIGGER TO X;
--  GRANT GRANT ANY PRIVILEGE TO X WITH ADMIN OPTION;
--  GRANT INSERT ANY TABLE TO X;
--  GRANT SELECT ANY TABLE TO X;


  TYPE TableDef                IS TABLE OF sys.all_tables%rowtype       INDEX BY BINARY_INTEGER;
  TYPE TableColumns            IS TABLE OF sys.all_tab_columns%rowtype  INDEX BY BINARY_INTEGER;
  TYPE TableIndexes            IS TABLE OF sys.all_indexes%rowtype      INDEX BY BINARY_INTEGER;
  TYPE TableIndexColumns       IS TABLE OF sys.all_ind_columns%rowtype  INDEX BY BINARY_INTEGER;
  TYPE TableConstraints        IS TABLE OF sys.all_constraints%rowtype  INDEX BY BINARY_INTEGER;
  TYPE TableConstraintColumns  IS TABLE OF sys.all_cons_columns%rowtype INDEX BY BINARY_INTEGER;
  TYPE TableTriggers           IS TABLE OF sys.all_triggers%rowtype     INDEX BY BINARY_INTEGER;
  TYPE TableTriggerColumns     IS TABLE OF sys.all_trigger_cols%rowtype INDEX BY BINARY_INTEGER;
  TYPE TableColumnComments     IS TABLE OF sys.all_col_comments%rowtype INDEX BY BINARY_INTEGER;
  TYPE TableComments           IS TABLE OF sys.all_tab_comments%rowtype INDEX BY BINARY_INTEGER;
  TYPE TableFConstraints       IS TABLE OF sys.all_constraints%rowtype  INDEX BY BINARY_INTEGER;
  TYPE TableFConstraintColumns IS TABLE OF sys.all_cons_columns%rowtype INDEX BY BINARY_INTEGER;
  TYPE TableSynonyms           IS TABLE OF sys.all_synonyms%rowtype     INDEX BY BINARY_INTEGER;
  TYPE TableGrants             IS TABLE OF sys.all_tab_privs%rowtype    INDEX BY BINARY_INTEGER;
  TYPE TableColumnGrants       IS TABLE OF sys.all_col_privs%rowtype    INDEX BY BINARY_INTEGER;

  -- Read the definition of table 'tn' from the database
  -- Unfortunately PL/SQL doesn't allow records with tables of records :-(
  -- Then only two parameters would be needed.
  PROCEDURE GetTableDefinition(tn       IN  VARCHAR2, 
                               tdef     OUT TableDef, 
                               tcol     OUT TableColumns,
                               tind     OUT TableIndexes,
                               tindcol  OUT TableIndexColumns,
                               tcon     OUT TableConstraints,
                               tconcol  OUT TableConstraintColumns,
                               ttrig    OUT TableTriggers,
                               ttrigcol OUT TableTriggerColumns,
                               tcolcom  OUT TableColumnComments,
                               tcom     OUT TableComments,
                               tfcon    OUT TableFConstraints,
                               tfconcol OUT TableFConstraintColumns,
                               tsyn     OUT TableSynonyms,
                               tgrnt    OUT TableGrants,
                               tcgrnt   OUT TableColumnGrants,
                               owner_id IN VARCHAR2 DEFAULT USER);


  -- Sets the name of the temporary table that will be created (default=tmp_tabman)
  PROCEDURE SetTemporaryTable(table_name VARCHAR2,
                              owner_id   VARCHAR2 DEFAULT USER);

  -- Sets and clears the output file 'filename'
  -- This file will contain the SQL statements used
  -- A NULL value (default) uses no output file.
  -- If Debug=TRUE, the debug info is also written to the file
  -- Make sure the path is uncluded in the utl_file_dir intialization parameter!
  PROCEDURE OutputFilename(filename VARCHAR2,debug BOOLEAN DEFAULT FALSE);

  -- Sets the name of the pipe where output is sent to.
  -- If pipe_name is NULL (default) then no output is sent through a pipe.
  -- If sql_stmts is TRUE, then all issued SQL statements are passed through the pipe,
  -- so that the consumer can writes these to a file or something.
  -- If debug_info is TRUE, then debug information is also sent through the pipe.
  -- A message has the following structure:
  --   message_type VARCHAR2(10)
  --   message_text VARCHAR2(4000)
  -- The following messages can be sent:
  --   * message_type='INFO'
  --     message_text=The text of what the procedure is doing
  --
  --   * message_type='ERROR'
  --     message_text=The error message
  --
  --   * message_type='WARNING'
  --     message_text=The warning message
  --
  --   * message_type='SQL'
  --     message_text=The SQL statement that's being executed
  --
  --   * message_type='DEBUG'
  --     message_text=The debug information
  --
  --   * message_type='END'  This is the last message that's sent through the pipe.
  --     message_text='SUCCESS', 'SUCCESS_WITH_WARNINGS' or 'ERROR'
  PROCEDURE SetPipeName(pipe_name  VARCHAR2,
                        sql_stmts  BOOLEAN DEFAULT FALSE, 
                        debug_info BOOLEAN DEFAULT FALSE);
                        
  -- Returns the current version of the package as a string
  FUNCTION GetVersion RETURN VARCHAR2;
  pragma restrict_references (GetVersion, WNDS);
  
  -- Returns the major, minor, patch, description release version as three out variables
  PROCEDURE GetRelease(major       OUT NUMBER, 
                       minor       OUT NUMBER,
                       patch       OUT NUMBER,
                       description OUT VARCHAR2);

  -- Executes the DDL statement 'stmt'
  PROCEDURE DDL(stmt IN VARCHAR2);

  -- Returns the row count of the table 'table_name'
  -- You can specify an optional 'where_clause' (with or without 'WHERE')
  FUNCTION CountTable(table_name   IN VARCHAR2, 
                      where_clause IN VARCHAR2 DEFAULT NULL,
                      owner_id     IN VARCHAR2 DEFAULT USER) RETURN NUMBER;

  -- Droppes the table 'table_name'
  PROCEDURE DropTable(table_name IN VARCHAR2,
                      owner_id   IN VARCHAR2 DEFAULT USER);

  -- Droppes the column 'cn' of table 'tn'
  PROCEDURE DropColumn(tn       IN VARCHAR2, 
                       cn       IN VARCHAR2,
                       owner_id IN VARCHAR2 DEFAULT USER);

  -- Droppes the constraint 'constraint_name'
  PROCEDURE DropConstraint(constraint_name IN VARCHAR2,
                           owner_id        IN VARCHAR2 DEFAULT USER);

  -- Droppes the index 'index_name'
  PROCEDURE DropIndex(index_name IN VARCHAR2,
                      owner_id   IN VARCHAR2 DEFAULT USER);

  -- Droppes the trigger 'trigger_name'
  PROCEDURE DropTrigger(trigger_name IN VARCHAR2,
                        owner_id     IN VARCHAR2 DEFAULT USER);

  -- Droppes the comment on table 'table_name'
  PROCEDURE DropTableComment(table_name IN VARCHAR2,
                             owner_id   IN VARCHAR2 DEFAULT USER);

  -- Droppes the comment on column 'column_name' of table 'table_name'
  PROCEDURE DropColumnComment(table_name  IN VARCHAR2,
                              column_name IN VARCHAR2,
                              owner_id    IN VARCHAR2 DEFAULT USER);

  -- Alters the table 'tn' with the new properties.
  -- Default values mean no change.
  -- Example: To change the initial extent of a table:
  --          TabMan.AlterTable('my_table',initial_extent=>10000);
  PROCEDURE AlterTable(tn              IN VARCHAR2,
  		       table_name      IN VARCHAR2 DEFAULT NULL,
  		       tablespace_name IN VARCHAR2 DEFAULT NULL,
  		       pct_free        IN NUMBER   DEFAULT NULL,
  		       pct_used        IN NUMBER   DEFAULT NULL, -- Use this for pct_threshold for Index-Organized tables
  		       ini_trans       IN NUMBER   DEFAULT NULL,
  		       max_trans       IN NUMBER   DEFAULT NULL,
  		       initial_extent  IN NUMBER   DEFAULT NULL,
  		       next_extent     IN NUMBER   DEFAULT NULL,
  		       min_extents     IN NUMBER   DEFAULT NULL,
  		       max_extents     IN NUMBER   DEFAULT NULL,
  		       pct_increase    IN NUMBER   DEFAULT NULL,
  		       freelists       IN NUMBER   DEFAULT NULL,
  		       freelist_groups IN NUMBER   DEFAULT NULL,
  		       owner_id        IN VARCHAR2 DEFAULT USER,
                       new_owner       IN VARCHAR2 DEFAULT NULL);
  		       
  -- Alters the tables that match the criteria of
  -- parameters 's_xxxxx' with the new properties.
  -- Default values mean no change.
  -- Example: To change move all tables from tablespace X to Y
  --          TabMan.AlterTables(s_tablespace_name=>'X',tablespace_name=>'Y');
  PROCEDURE AlterTables(tn                IN VARCHAR2 DEFAULT NULL,
                        s_table_name      IN VARCHAR2 DEFAULT NULL,
                        s_tablespace_name IN VARCHAR2 DEFAULT NULL,
                        s_pct_free        IN NUMBER   DEFAULT NULL,
                        s_pct_used        IN NUMBER   DEFAULT NULL,
                        s_ini_trans       IN NUMBER   DEFAULT NULL,
                        s_max_trans       IN NUMBER   DEFAULT NULL,
                        s_initial_extent  IN NUMBER   DEFAULT NULL,
                        s_next_extent     IN NUMBER   DEFAULT NULL,
                        s_min_extents     IN NUMBER   DEFAULT NULL,
                        s_max_extents     IN NUMBER   DEFAULT NULL,
                        s_pct_increase    IN NUMBER   DEFAULT NULL,
                        s_freelists       IN NUMBER   DEFAULT NULL,
                        s_freelist_groups IN NUMBER   DEFAULT NULL,
                        tablespace_name   IN VARCHAR2 DEFAULT NULL,
                        pct_free          IN NUMBER   DEFAULT NULL,
                        pct_used          IN NUMBER   DEFAULT NULL, -- Use this for pct_threshold for Index-Organized tables
                        ini_trans         IN NUMBER   DEFAULT NULL,
                        max_trans         IN NUMBER   DEFAULT NULL,
                        initial_extent    IN NUMBER   DEFAULT NULL,
                        next_extent       IN NUMBER   DEFAULT NULL,
                        min_extents       IN NUMBER   DEFAULT NULL,
                        max_extents       IN NUMBER   DEFAULT NULL,
                        pct_increase      IN NUMBER   DEFAULT NULL,
                        freelists         IN NUMBER   DEFAULT NULL,
                        freelist_groups   IN NUMBER   DEFAULT NULL,
                        s_owner_id        IN VARCHAR2 DEFAULT USER,
                        owner_id          IN VARCHAR2 DEFAULT NULL);

  -- Alters the index 'ind' with the new properties.
  -- Default values mean no change.
  -- Example: To change the initial extent of an index:
  --          TabMan.AlterIndex('my_index',initial_extent=>10000);
  PROCEDURE AlterIndex(ind             IN VARCHAR2,
  		       index_name      IN VARCHAR2 DEFAULT NULL,
  		       index_type      IN VARCHAR2 DEFAULT NULL,
 	               tablespace_name IN VARCHAR2 DEFAULT NULL,
 	               uniqueness      IN BOOLEAN  DEFAULT NULL,
  		       pct_free        IN NUMBER   DEFAULT NULL,
  		       ini_trans       IN NUMBER   DEFAULT NULL,
  		       max_trans       IN NUMBER   DEFAULT NULL,
  		       initial_extent  IN NUMBER   DEFAULT NULL,
  		       next_extent     IN NUMBER   DEFAULT NULL,
  		       min_extents     IN NUMBER   DEFAULT NULL,
  		       max_extents     IN NUMBER   DEFAULT NULL,
  		       pct_increase    IN NUMBER   DEFAULT NULL,
  		       freelists       IN NUMBER   DEFAULT NULL,
  		       freelist_groups IN NUMBER   DEFAULT NULL,
                       owner_id        IN VARCHAR2 DEFAULT USER,
                       new_owner_id    IN VARCHAR2 DEFAULT NULL);

  -- Alters the index 'ind' with the new properties.
  -- Default values mean no change.
  -- This procedure allows you to change the columns of the index also
  -- by specifying a table of TableIndexColumns
  PROCEDURE AlterIndex(ind             IN VARCHAR2,
  		       index_name      IN VARCHAR2 DEFAULT NULL,
  		       index_type      IN VARCHAR2 DEFAULT NULL, -- NORMAL, BITMAP, NORMAL/REV, BITMAP/REV
 	               tablespace_name IN VARCHAR2 DEFAULT NULL,
 	               uniqueness      IN BOOLEAN  DEFAULT NULL,
  		       pct_free        IN NUMBER   DEFAULT NULL,
  		       ini_trans       IN NUMBER   DEFAULT NULL,
  		       max_trans       IN NUMBER   DEFAULT NULL,
  		       initial_extent  IN NUMBER   DEFAULT NULL,
  		       next_extent     IN NUMBER   DEFAULT NULL,
  		       min_extents     IN NUMBER   DEFAULT NULL,
  		       max_extents     IN NUMBER   DEFAULT NULL,
  		       pct_increase    IN NUMBER   DEFAULT NULL,
  		       freelists       IN NUMBER   DEFAULT NULL,
  		       freelist_groups IN NUMBER   DEFAULT NULL,
  		       index_columns   IN TableIndexColumns,
                       owner_id        IN VARCHAR2 DEFAULT USER,
                       new_owner       IN VARCHAR2 DEFAULT NULL);
  		       
  -- Alters the indexes that match the criteria of
  -- parameters 's_xxxxx' with the new properties.
  -- Default values mean no change.
  -- Example: To change move all indexes from tablespace X to Y
  --          TabMan.AlterIndexes(s_tablespace_name=>'X',tablespace_name=>'Y');
  PROCEDURE AlterIndexes(s_index_name      IN VARCHAR2 DEFAULT NULL,
                         s_index_type      IN VARCHAR2 DEFAULT NULL, -- NORMAL, BITMAP, NORMAL/REV, BITMAP/REV
                         s_tablespace_name IN VARCHAR2 DEFAULT NULL,
                         s_uniqueness      IN BOOLEAN  DEFAULT NULL,
                         s_pct_free        IN NUMBER   DEFAULT NULL,
                         s_ini_trans       IN NUMBER   DEFAULT NULL,
                         s_max_trans       IN NUMBER   DEFAULT NULL,
                         s_initial_extent  IN NUMBER   DEFAULT NULL,
                         s_next_extent     IN NUMBER   DEFAULT NULL,
                         s_min_extents     IN NUMBER   DEFAULT NULL,
                         s_max_extents     IN NUMBER   DEFAULT NULL,
                         s_pct_increase    IN NUMBER   DEFAULT NULL,
                         s_freelists       IN NUMBER   DEFAULT NULL,
                         s_freelist_groups IN NUMBER   DEFAULT NULL,
                         index_type        IN VARCHAR2 DEFAULT NULL, -- NORMAL, BITMAP, NORMAL/REV, BITMAP/REV
                         tablespace_name   IN VARCHAR2 DEFAULT NULL,
                         uniqueness        IN BOOLEAN  DEFAULT NULL,
                         pct_free          IN NUMBER   DEFAULT NULL,
                         ini_trans         IN NUMBER   DEFAULT NULL,
                         max_trans         IN NUMBER   DEFAULT NULL,
                         initial_extent    IN NUMBER   DEFAULT NULL,
                         next_extent       IN NUMBER   DEFAULT NULL,
                         min_extents       IN NUMBER   DEFAULT NULL,
                         max_extents       IN NUMBER   DEFAULT NULL,
                         pct_increase      IN NUMBER   DEFAULT NULL,
                         freelists         IN NUMBER   DEFAULT NULL,
                         freelist_groups   IN NUMBER   DEFAULT NULL,
                         s_owner_id        IN VARCHAR2 DEFAULT USER,
                         owner_id          IN VARCHAR2 DEFAULT NULL);

  -- Alters the constraint 'cn' with the new properties.
  -- Default values mean no change.
  -- Example: To change the status of a constraint: TabMan.AlterConstraint('qty_cc',status=>'DISABLED');
  PROCEDURE AlterConstraint(cn                IN VARCHAR2,
       		            constraint_name   IN VARCHAR2 DEFAULT NULL,
                            constraint_type   IN VARCHAR2 DEFAULT NULL, -- P, C, R, U
                            table_name        IN VARCHAR2 DEFAULT NULL,
                            search_condition  IN VARCHAR2 DEFAULT NULL,
                            r_owner           IN VARCHAR2 DEFAULT NULL,
                            r_constraint_name IN VARCHAR2 DEFAULT NULL,
                            delete_rule       IN VARCHAR2 DEFAULT NULL, -- CASCADE, NO_ACTION
                            status            IN VARCHAR2 DEFAULT NULL, -- ENABLED, DISABLED
                            deferrable        IN VARCHAR2 DEFAULT NULL, -- DEFERRABLE, NOT DEFERRABLE
                            deferred          IN VARCHAR2 DEFAULT NULL, -- IMMEDIATE, DEFERRED
                            validated         IN VARCHAR2 DEFAULT NULL, -- VALIDATED, NOT VALIDATED
                            owner_id          IN VARCHAR2 DEFAULT USER,
                            new_owner_id      IN VARCHAR2 DEFAULT NULL);

  -- Alters the constraint 'cn' with the new properties.
  -- Default values mean no change.
  -- You can change the columns where the constraint is on with 'cons_columns'
  -- Columns must be in order in the PL/SQL table !! (position is ignored)
  -- Example: To change the status of a constraint: TabMan.AlterConstraint('qty_cc',status=>'DISABLED');
  PROCEDURE AlterConstraint(cn                IN VARCHAR2,
                            cons_columns      IN TableConstraintColumns,
       		            constraint_name   IN VARCHAR2 DEFAULT NULL,
                            constraint_type   IN VARCHAR2 DEFAULT NULL, -- P, C, R, U
                            table_name        IN VARCHAR2 DEFAULT NULL,
                            search_condition  IN VARCHAR2 DEFAULT NULL,
                            r_owner           IN VARCHAR2 DEFAULT NULL,
                            r_constraint_name IN VARCHAR2 DEFAULT NULL,
                            delete_rule       IN VARCHAR2 DEFAULT NULL, -- CASCADE, NO_ACTION
                            status            IN VARCHAR2 DEFAULT NULL, -- ENABLED, DISABLED
                            deferrable        IN VARCHAR2 DEFAULT NULL, -- DEFERRABLE, NOT DEFERRABLE
                            deferred          IN VARCHAR2 DEFAULT NULL, -- IMMEDIATE, DEFERRED
                            validated         IN VARCHAR2 DEFAULT NULL, -- VALIDATED, NOT VALIDATED
                            owner_id          IN VARCHAR2 DEFAULT USER,
                            new_owner_id      IN VARCHAR2 DEFAULT NULL);

  -- **********************************************************
  -- Here follow some handy ultility procedures...
  -- **********************************************************
  
  -- Searches tables for a specified text string and stores the result in a table
  -- It searches only in CHAR, VARCHAR2, DATE (fmt=dd/mm/yyyy hh24:mi:ss) and NUMBER columns.
  -- The result table has the follwing colums:
  --  - owner       VARCHAR2(30)  : The owner of the table
  --  - table_name  VARCHAR2(30)  : The name of the table
  --  - column_name VARCHAR2(30)  : The name of the column
  --  - row_id      ROWID         : The rowid of the found row
  --  - text        VARCHAR2(4000): The matching text
  -- Example: TabMan.Grep('TABMAN') finds the text TABMAN in all columns of all tables of the current schema.
  PROCEDURE Grep(what     IN VARCHAR2,               -- The text to find (may contain wildcards)
                 tn       IN VARCHAR2 DEFAULT NULL,  -- The tables to search (default=all tables in the current schema)
                 cn       IN VARCHAR2 DEFAULT NULL,  -- The columns to search (default=all columns in the tables)
                 owner_id IN VARCHAR2 DEFAULT USER,  -- The owner of the tables (default=current user)
		 casesens IN BOOLEAN  DEFAULT FALSE, -- Search case sensitive
                 result   IN VARCHAR2 DEFAULT 'TABMAN_GREP_RESULTS'); -- The name of the result table (created or truncated by grep)


END;
/
show err

@@tabmanb.sql


