CREATE OR REPLACE PACKAGE BODY TabMan AS

  g_major       NUMBER      :=2;
  g_minor       NUMBER      :=0;
  g_patch       NUMBER      :=0;
  g_description VARCHAR2(40):='Preview 7';
  
  TYPE LogTable IS TABLE OF VARCHAR2(2000) INDEX BY BINARY_INTEGER;
  temporary_table       VARCHAR2(30) :='tmp_tabman';
  temporary_table_owner VARCHAR2(30) :=USER;
  output_file           VARCHAR2(512):=NULL;
  output_file_dir       VARCHAR2(512):=NULL;
  log_table             LogTable;
  output_pipe           VARCHAR2(30) :=NULL;
  output_sql_stmts      BOOLEAN      :=FALSE;
  output_debug_info     BOOLEAN      :=FALSE;
  logresult             INTEGER      :=0;
  file_debug            BOOLEAN      :=FALSE;

  FUNCTION GetVersion RETURN VARCHAR2 IS
  BEGIN
    return 'TabMan: Release'||' '||to_char(g_major)||'.'||to_char(g_minor)||
           '.'||to_char(g_patch)||' '||g_description||' - Production on '||to_char(sysdate,'Dy Mon dd hh24:mi:ss yyyy');
  END;
  
  PROCEDURE GetRelease(major OUT NUMBER, 
                       minor OUT NUMBER,
                       patch OUT NUMBER,
                       description OUT VARCHAR2) IS
  BEGIN
    major:=g_major;
    minor:=g_minor;
    patch:=g_patch;
    description:=g_description;
  END;

  PROCEDURE LogToFile(str IN VARCHAR2) IS
  f utl_file.file_type;
  BEGIN
    BEGIN
      f:=utl_file.fopen(output_file_dir,output_file,'A');
    EXCEPTION WHEN OTHERS THEN
      f:=utl_file.fopen(output_file_dir,output_file,'W');
    END;
    utl_file.put_line(f,str);
    utl_file.fclose(f);
  END;
  
  PROCEDURE WorkingLog(str VARCHAR2) IS
  x number;
  BEGIN
    if output_pipe is not null then
      dbms_pipe.pack_message('INFO');
      dbms_pipe.pack_message(str);
      x:=dbms_pipe.send_message(output_pipe);
    end if;
    if output_file is not null then
      LogToFile('prompt '||str);
    end if;
  END;

  PROCEDURE WarningLog(str VARCHAR2) IS
  x number;
  BEGIN
    if output_pipe is not null then
      dbms_pipe.pack_message('WARNING');
      dbms_pipe.pack_message(str);
      x:=dbms_pipe.send_message(output_pipe);
      if logresult=0 then
        logresult:=1;
      end if;
    end if;
  END;

  PROCEDURE ErrorLog(str VARCHAR2) IS
  x number;
  BEGIN
    if output_pipe is not null then
      dbms_pipe.pack_message('ERROR');
      dbms_pipe.pack_message(str);
      x:=dbms_pipe.send_message(output_pipe);
      if logresult<2 then
        logresult:=2;
      end if;
    else
      raise_application_error(-20000,str,TRUE);
    end if;
  END;
  
  PROCEDURE ApplicationError(errno NUMBER,errtxt VARCHAR) IS
  BEGIN
    if output_pipe is not null then
      ErrorLog(errtxt);
    else
      raise_application_error(errno,errtxt,TRUE);
    end if;
  END;

  PROCEDURE Debug(str VARCHAR2) IS
  x number;
  BEGIN
    if output_pipe is not null and output_debug_info then
      dbms_pipe.pack_message('DEBUG');
      dbms_pipe.pack_message(str);
      x:=dbms_pipe.send_message(output_pipe);
    end if;
    if output_file is not null and file_debug then
      LogToFile('prompt '||str);
    end if;
  END;

  PROCEDURE OutputSQL(str VARCHAR2) IS
  x number;
  BEGIN
    if output_pipe is not null and output_sql_stmts then
      dbms_pipe.pack_message('SQL');
      dbms_pipe.pack_message(str);
      x:=dbms_pipe.send_message(output_pipe);
    end if;
    if output_file is not null then
      LogToFile(str);
      LogToFile('/');
    end if;
  END;

  PROCEDURE ClearLog IS 
  BEGIN
    WorkingLog('Initializing...');
    if output_pipe is null then
      log_table.delete;
    else
      dbms_pipe.purge(output_pipe);
    end if;
    logresult:=0;
  END;
  
  PROCEDURE AddToLog(str VARCHAR2) IS
  BEGIN
    if output_pipe is null then
      log_table(nvl(log_table.last,0)+1):=str;
    else
      WarningLog(str);
    end if;
  END;
  
  PROCEDURE PrintLog IS
  str VARCHAR2(32000);
  x number;
  BEGIN
    if output_pipe is null then
      if log_table.last is not null then
        str:='Operation finished successfully with following warnings:';
        for x in log_table.first..log_table.last loop
          if log_table.exists(x) then
            str:=str||CHR(10)||log_table(x);
          end if;
        end loop;
        raise_application_error(-20006,str);
      end if;
    else
      if logresult=0 then
        dbms_pipe.pack_message('END');
        dbms_pipe.pack_message('SUCCESS');
        x:=dbms_pipe.send_message(output_pipe);
      elsif logresult=1 then
        dbms_pipe.pack_message('END');
        dbms_pipe.pack_message('SUCCESS_WITH_WARNINGS');
        x:=dbms_pipe.send_message(output_pipe);
      elsif logresult=2 then
        dbms_pipe.pack_message('END');
        dbms_pipe.pack_message('ERROR');
        x:=dbms_pipe.send_message(output_pipe);
      else
        dbms_pipe.pack_message('END');
        dbms_pipe.pack_message('UNKNOWN');
        x:=dbms_pipe.send_message(output_pipe);
      end if;        
    end if;
  END;    

  PROCEDURE SetPipeName(pipe_name VARCHAR2,
                        sql_stmts BOOLEAN DEFAULT FALSE, 
                        debug_info BOOLEAN DEFAULT FALSE) IS
  BEGIN
    output_pipe:=pipe_name;
    output_sql_stmts:=sql_stmts;
    output_debug_info:=debug_info;
    if output_pipe is not null then
      dbms_pipe.purge(output_pipe);
    end if;
  END;
                          
  PROCEDURE DDL(stmt IN VARCHAR2) IS
  source_cursor INTEGER;
  ignore        INTEGER;
  BEGIN
    OutputSQL(stmt);
    source_cursor := dbms_sql.open_cursor;
    dbms_sql.parse(source_cursor,stmt,dbms_sql.native);
    ignore := dbms_sql.execute(source_cursor);
    dbms_sql.close_cursor(source_cursor);
  EXCEPTION WHEN OTHERS THEN
    IF dbms_sql.is_open(source_cursor) THEN
      dbms_sql.close_cursor(source_cursor);
    END IF;
    RAISE;
  END;

  PROCEDURE OutputFilename(filename VARCHAR2,debug BOOLEAN DEFAULT FALSE) IS
  f utl_file.file_type;
  pos number;
  BEGIN
    pos:=instr(filename,'/',-1);
    if pos=0 then
      pos:=instr(filename,'\',-1);
    end if;
    if pos=0 then
      raise_application_error(-20004,'You must specify a destination directory');
      return;
    end if;
    output_file:=substr(filename,pos+1);
    output_file_dir:=substr(filename,1,pos-1);
    f:=utl_file.fopen(output_file_dir,output_file,'W');
    utl_file.fclose(f);
    file_debug:=debug;
  EXCEPTION WHEN OTHERS THEN
    output_file:=null;
    raise;
  END;

  PROCEDURE GetTableDef(tn IN VARCHAR2,
                        tdef IN OUT TableDef,
                        owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_tables 
              where table_name = upper(tn) 
              and owner=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableDef');
    open c;
    loop
      fetch c into tdef(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableDef');
  END;

  PROCEDURE GetTableColumns(tn IN VARCHAR2,
                            tcol IN OUT TableColumns,
                            owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_tab_columns 
              where table_name = upper(tn) 
              and owner= upper(owner_id)
              order by column_id;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableColumns');
    open c;
    loop
      fetch c into tcol(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableColumns');
  END;

  PROCEDURE GetTableIndexes(tn IN VARCHAR2,
                            tind IN OUT TableIndexes,
                            owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_indexes 
              where table_name = upper(tn)
              and owner=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableIndexes');
    open c;
    loop
      fetch c into tind(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableIndexes');
  END;

  PROCEDURE GetTableIndexColumns(tn IN VARCHAR2,
                                 tindcol IN OUT TableIndexColumns,
                                 owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_ind_columns 
              where table_name = upper(tn) 
              and table_owner=upper(owner_id)
              order by index_name,column_position;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableIndexColumns');
    open c;
    loop
      fetch c into tindcol(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableIndexColumns');
  END;

  PROCEDURE GetIndexDef(ind      IN     VARCHAR2,
                        tind     IN OUT TableIndexes,
                        owner_id IN     VARCHAR2) IS
  cursor c is select * from sys.all_indexes 
              where index_name = upper(ind)
              and owner=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetIndexDef');
    open c;
    loop
      fetch c into tind(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetIndexDef');
  END;

  PROCEDURE GetConstraintDef(cn       IN     VARCHAR2,
                             tcon     IN OUT TableConstraints,
                             owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_constraints
              where constraint_name = upper(cn)
              and owner=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetConstraintDef');
    open c;
    loop
      fetch c into tcon(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetConstraintDef');
  END;

  PROCEDURE GetIndexColumns(ind IN VARCHAR2,
                            tindcol IN OUT TableIndexColumns,
                            owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_ind_columns 
              where index_name = upper(ind) 
              and index_owner=upper(owner_id)
              order by index_name,column_position;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetIndexColumns');
    open c;
    loop
      fetch c into tindcol(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetIndexColumns');
  END;

  PROCEDURE GetTableConstraints(tn IN VARCHAR2,
                                tcon IN OUT TableConstraints,
                                owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_constraints 
              where table_name = upper(tn) 
              and owner=upper(owner_id)
              and constraint_name not like 'SYS\_%' ESCAPE '\';
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableConstraints');
    open c;
    loop
      fetch c into tcon(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableConstraints');
  END;

  PROCEDURE GetTableConstraintColumns(tn IN VARCHAR2,
                                      tconcol IN OUT TableConstraintColumns,
                                      owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_cons_columns 
              where table_name = upper(tn) 
              and owner=upper(owner_id)
              and constraint_name not like 'SYS\_%' ESCAPE '\'
              order by position;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableConstraintColumns');
    open c;
    loop
      fetch c into tconcol(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableConstraintColumns');
  END;

  PROCEDURE GetConstraintColumns(cn IN VARCHAR2,
                                 tconcol IN OUT TableConstraintColumns,
                                 owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_cons_columns 
              where constraint_name = upper(cn) 
              and owner=upper(owner_id)
              order by position;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetConstraintColumns');
    open c;
    loop
      fetch c into tconcol(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetConstraintColumns');
  END;

  PROCEDURE GetTableTriggers(tn IN VARCHAR2,
                             ttrig IN OUT TableTriggers,
                             owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_triggers 
              where table_name = upper(tn)
              and owner=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableTriggers');
    open c;
    loop
      fetch c into ttrig(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableTriggers');
  END;

  PROCEDURE GetTableColumnComments(tn IN VARCHAR2,
                                   tcolcom IN OUT TableColumnComments,
                                   owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_col_comments 
              where table_name = upper(tn) 
              and owner=upper(owner_id)
              and comments is not null;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableColumnComments');
    open c;
    loop
      fetch c into tcolcom(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableColumnComments');
  END;

  PROCEDURE GetTableComments(tn IN VARCHAR2,
                             tcom IN OUT TableComments,
                             owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_tab_comments 
              where table_name = upper(tn) 
              and owner=upper(owner_id)
              and comments is not null;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableComments');
    open c;
    loop
      fetch c into tcom(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableComments');
  END;

  PROCEDURE GetTableFConstraints(tn IN VARCHAR2,
                                 tfcon IN OUT TableFConstraints,
                                 owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_constraints 
              where (r_constraint_name,r_owner) in 
               (select constraint_name,owner
                from sys.all_constraints 
                where table_name=upper(tn)
                and owner=upper(owner_id))
              and constraint_name not like 'SYS\_%' ESCAPE '\' 
              order by constraint_type;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableFConstraints');
    open c;
    loop
      fetch c into tfcon(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableFConstraints');
  END;

  PROCEDURE GetTableFConstraintColumns(tn IN VARCHAR2,
                                       tfconcol IN OUT TableFConstraintColumns,
                                       owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_cons_columns 
              where (constraint_name,owner) in (
                select constraint_name,owner from sys.all_constraints 
                where (r_constraint_name,r_owner) in (
                  select constraint_name,owner from sys.all_constraints 
                  where table_name=upper(tn)
                  and owner=upper(owner_id)) 
                and constraint_name not like 'SYS\_%' ESCAPE '\')
              order by position;
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableFConstraintColumns');
    open c;
    loop
      fetch c into tfconcol(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableFConstraintColumns');
  END;

  PROCEDURE GetTableSynonyms(tn       IN VARCHAR2,
                             tsyn     IN OUT TableSynonyms,
                             owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_synonyms
              where table_name=upper(tn)
                and table_owner=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableSynonyms');
    open c;
    loop
      fetch c into tsyn(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableSynonyms');
  END;

  PROCEDURE GetTableGrants(tn       IN VARCHAR2,
                           tgrnt    IN OUT TableGrants,
                           owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_tab_privs
              where table_name=upper(tn)
                and table_schema=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableGrants');
    open c;
    loop
      fetch c into tgrnt(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableGrants');
  END;

  PROCEDURE GetTableColumnGrants(tn       IN VARCHAR2,
                                 tcgrnt   IN OUT TableColumnGrants,
                                 owner_id IN VARCHAR2) IS
  cursor c is select * from sys.all_col_privs
              where table_name=upper(tn)
                and table_schema=upper(owner_id);
  x pls_integer:=1;
  BEGIN
    Debug('Start GetTableColumnGrants');
    open c;
    loop
      fetch c into tcgrnt(x);
      exit when c%notfound;
      x:=x+1;
    end loop;
    close c;
    Debug('End GetTableColumnGrants');
  END;

  PROCEDURE SetTemporaryTable(table_name VARCHAR2,
                              owner_id IN VARCHAR2 DEFAULT USER) IS
  BEGIN
    temporary_table:=table_name;
    temporary_table_owner:=owner_id;
  END;
  
  PROCEDURE DropTemporaryTable IS
  BEGIN
    Debug('Start DropTemporaryTable');
    DDL('drop table "'||temporary_table_owner||'".'||temporary_table);
    Debug('End DropTemporaryTable');
  EXCEPTION WHEN OTHERS THEN
    Debug('DropTemporaryTable: '||sqlerrm);
    NULL;
  END;
  
  PROCEDURE RemoveColumn(cn       IN VARCHAR2,
                         tcol     IN OUT TableColumns,
                         tind     IN OUT TableIndexes,
                         tindcol  IN OUT TableIndexColumns,
                         tcon     IN OUT TableConstraints,
                         tconcol  IN OUT TableConstraintColumns,
                         ttrig    IN OUT TableTriggers,
                         ttrigcol IN OUT TableTriggerColumns,
                         tcolcom  IN OUT TableColumnComments,
                         tcom     IN OUT TableComments,
                         tfcon    IN OUT TableFConstraints,
                         tfconcol IN OUT TableFConstraintColumns) IS
  found BOOLEAN;                         
  BEGIN
    Debug('Start RemoveColumn');
    WorkingLog('Removing column '||upper(cn)||' from table definition');
    for x in nvl(tcol.first,0)..nvl(tcol.last,0) loop
      if tcol.exists(x) and tcol(x).column_name=upper(cn) then
        tcol.Delete(x);
        exit;
      end if;
    end loop;
    Debug('RemoveColumn 1');
    for x in nvl(tindcol.first,0)..nvl(tindcol.last,0) loop
      if tindcol.exists(x) and tindcol(x).column_name=upper(cn) then
        tindcol.Delete(x);
      end if;
    end loop;
    Debug('RemoveColumn 2');
    for x in nvl(tind.first,0)..nvl(tind.last,0) loop
      found:=false;
      if tind.exists(x) then
        for y in nvl(tindcol.first,0)..nvl(tindcol.last,0) loop
          if tindcol.exists(y) and tind(x).index_name=tindcol(y).index_name then
            found:=TRUE;
            exit;
          end if;
        end loop;
        if not found then
          tind.Delete(x);
        end if;
      end if;
    end loop;
    Debug('RemoveColumn 3');
    for x in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
      if tconcol.exists(x) and tconcol(x).column_name=upper(cn) then
        for y in nvl(tfcon.first,0)..nvl(tfcon.last,0) loop
          if tfcon.exists(y) and tfcon(y).r_constraint_name=tconcol(x).constraint_name then
            for z in nvl(tfconcol.first,0)..nvl(tfconcol.last,0) loop
              if tfconcol.exists(z) and tfconcol(z).constraint_name=tfcon(y).constraint_name and tconcol(x).position=tfconcol(z).position then
                tfconcol.Delete(z);
                exit;
              end if;
            end loop;
          end if;
        end loop;
        tconcol.Delete(x);
      end if;
    end loop;
    Debug('RemoveColumn 4');
    for x in nvl(tcon.first,0)..nvl(tcon.last,0) loop
      if tcon.exists(x) then
        found:=false;
        for y in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
          if tconcol.exists(y) and tconcol(y).constraint_name=tcon(x).constraint_name then
            found:=true;
            exit;
          end if;
        end loop;
        if not found then
          tcon.Delete(x);
        end if;
      end if;
    end loop;
    Debug('RemoveColumn 5');
    for x in nvl(tfcon.first,0)..nvl(tfcon.last,0) loop
      found:=false;
      if tfcon.exists(x) then
        for y in nvl(tcon.first,0)..nvl(tcon.last,0) loop
          if tcon.exists(y) and tcon(y).constraint_name=tfcon(x).r_constraint_name then
            found:=true;
            exit;
          end if;
        end loop;
        if not found then
          tfcon.Delete(x);
        end if;
      end if;
    end loop;
    Debug('RemoveColumn 6');
    for x in nvl(tfcon.first,0)..nvl(tfcon.last,0) loop
      if tfcon.exists(x) and tfcon(x).constraint_type!='C' then
        found:=false;
        for y in nvl(tfconcol.first,0)..nvl(tfconcol.last,0) loop
          if tfconcol.exists(y) and tfconcol(y).constraint_name=tfcon(x).constraint_name then
            found:=true;
            exit;
          end if;
        end loop;
        if not found then
          tfcon.Delete(x);
        end if;
      end if;
    end loop;
    Debug('RemoveColumn 7');
    for x in nvl(tcolcom.first,0)..nvl(tcolcom.last,0) loop
      if tcolcom.exists(x) and tcolcom(x).column_name=upper(cn) then
        tcolcom.Delete(x);
        exit;
      end if;
    end loop;
    Debug('RemoveColumn 8');
    for x in nvl(ttrigcol.first,0)..nvl(ttrigcol.last,0) loop
      if ttrigcol.exists(x) and ttrigcol(x).column_name=upper(cn) then
        for y in nvl(ttrig.first,0)..nvl(ttrig.last,0) loop
          if ttrig.exists(y) and ttrigcol(x).trigger_name=ttrig(y).trigger_name then
            ttrig.Delete(y);
            exit;
          end if;
        end loop;
      end if;
    end loop;
    Debug('End RemoveColumn');
  END;

  PROCEDURE CopyTable(source          VARCHAR2,
                      target          VARCHAR2,
                      notcolumn       VARCHAR2,
                      nottcolumn      VARCHAR2,
                      except          BOOLEAN,
                      docolumn        VARCHAR2,
                      conv_function   VARCHAR2,
                      tcol            TableColumns,
                      source_owner_id VARCHAR2,
                      target_owner_id VARCHAR2) IS
  str       VARCHAR2(32000);
  scstr     VARCHAR2(32000);
  tcstr     VARCHAR2(32000);
  tcstr2    VARCHAR2(32000);
  second    BOOLEAN:=false;
  datatype  VARCHAR2(30);
  column    VARCHAR2(30);
  precision NUMBER;
  scale     NUMBER;
  cursor c is select column_name from sys.all_tab_columns where table_name=upper(source) and (notcolumn is null or column_name!=upper(notcolumn)) and owner=upper(source_owner_id)order by column_id;
  BEGIN
    Debug('Start CopyTable');
    WorkingLog('Copying data from '||upper(source_owner_id)||'.'||upper(source)||' to '||upper(target_owner_id)||'.'||upper(target));
    str:='declare'||chr(10)||'  tm_counter NUMBER:=0; '||chr(10)||'  ';
    for x in nvl(tcol.first,0)..nvl(tcol.last,0) loop
      if tcol.exists(x) then
        column:=tcol(x).column_name;
        if nottcolumn is null or upper(column)!=upper(nottcolumn) then
          str:=str||substr(column||'_',1,30)||' ';
          datatype:=tcol(x).data_type;
          str:=str||datatype;
          if datatype in ('VARCHAR2','NVARCHAR2','RAW','CHAR','NCHAR') then
            str:=str||'('||to_char(tcol(x).data_length)||')';
          elsif datatype='NUMBER' then
            precision:=tcol(x).data_precision;
            scale:=tcol(x).data_scale;
            if precision is not null then
              str:=str||'('||to_char(precision);
              if scale is not null then
                str:=str||','||to_char(scale);
              end if;
              str:=str||')';
            end if;
          end if;
          str:=str||'; '||chr(10)||'  ';
        end if;
      end if;
    end loop;
    Debug('CopyTable 1');
    str:=str||'cursor c is select ';
    second:=false;
    scstr:=' ';
    for x in c loop
      if second then
        scstr:=scstr||',';
      else
        second:=TRUE;
      end if;
      if docolumn is not null and x.column_name=upper(docolumn) then
        scstr:=scstr||chr(10)||'             '||conv_function;
      else
        scstr:=scstr||chr(10)||'             '||source||'.'||x.column_name;
      end if;
    end loop;
    second:=false;
    tcstr:=' ';
    tcstr2:=' ';
    Debug('CopyTable 2');
    for x in nvl(tcol.first,0)..nvl(tcol.last,0) loop
      if tcol.exists(x) then
        column:=tcol(x).column_name;
        if nottcolumn is null or upper(column)!=upper(nottcolumn) then
          if second then
            tcstr:=tcstr||',';
            tcstr2:=tcstr2||',';
          else
            second:=TRUE;
          end if;
          tcstr:=tcstr||chr(10)||'             '||substr(column||'_',1,30);
          tcstr2:=tcstr2||chr(10)||'             '||column;
        end if;
      end if;
    end loop;
    Debug('CopyTable 3');
    str:=str||scstr||' from "'||source_owner_id||'".'||source||';'||chr(10);
    str:=str||'begin'||chr(10);
    str:=str||'  open c;'||chr(10);
    str:=str||'  loop'||chr(10);
    str:=str||'    begin'||chr(10);
    str:=str||'      fetch c into '||tcstr||';'||chr(10);
    str:=str||'      exit when c%notfound;'||chr(10);
    str:=str||'      insert into "'||target_owner_id||'".'||target||' ('||tcstr2||')'||chr(10);
    str:=str||'      values ('||tcstr||');'||chr(10);
    str:=str||'      tm_counter:=tm_counter+1;'||chr(10);
    str:=str||'      if tm_counter>1000 then'||chr(10);
    str:=str||'        commit;'||chr(10);
    str:=str||'        tm_counter:=0;'||chr(10);
    str:=str||'      end if;'||chr(10);
    if except then
      str:=str||'    exception when others then'||chr(10);
      str:=str||'      null;'||chr(10);
    end if;
    str:=str||'    end;'||chr(10);
    str:=str||'  end loop;'||chr(10);
    str:=str||'  close c;'||chr(10);
    str:=str||'end;';
    DDL(str);
    Debug('End CopyTable');
  END;
    
  PROCEDURE CreateTable(tn       IN     VARCHAR2,
                        tdef     IN OUT TableDef,
                        tcol     IN     TableColumns,
                        owner_id IN     VARCHAR2,
                        tind     IN OUT TableIndexes,
                        tindcol  IN OUT TableIndexColumns,
                        tcon     IN OUT TableConstraints,
                        tconcol  IN OUT TableConstraintColumns,
                        orig_tn  IN     VARCHAR2) IS
  str       VARCHAR2(32000);
  def       VARCHAR2(32000);
  second    BOOLEAN:=false;
  datatype  VARCHAR2(30);
  column    VARCHAR2(30);
  precision NUMBER;
  scale     NUMBER;
  cn        VARCHAR2(30);
  tsn       VARCHAR2(30);
  ipct      NUMBER;
  ic        VARCHAR2(30);
  BEGIN
    Debug('Start CreateTable');
    WorkingLog('Creating table '||upper(owner_id)||'.'||upper(tn));
    str:='create table "'||owner_id||'".'||tn||' ('||chr(10)||' ';
    for x in nvl(tcol.first,0)..nvl(tcol.last,0) loop
      if tcol.exists(x) then
        if second then
          str:=str||','||chr(10)||' ';
        else
          second:=true;
        end if;
        column:=tcol(x).column_name;
        str:=str||rpad(column,30)||' ';
        datatype:=tcol(x).data_type;
        str:=str||datatype;
        if datatype in ('VARCHAR2','NVARCHAR2','RAW','CHAR','NCHAR') then
          str:=str||'('||to_char(tcol(x).data_length)||')';
        elsif datatype='NUMBER' then
          precision:=tcol(x).data_precision;
          scale:=tcol(x).data_scale;
          if precision is not null then
            str:=str||'('||to_char(precision);
            if scale is not null then
              str:=str||','||to_char(scale);
            end if;
            str:=str||')';
          elsif precision is null and scale is not null then
            str:=str||'(38)';
          end if;
        end if;
        def:=tcol(x).data_default;
        if def is not null then
          if datatype in ('VARCHAR2','NVARCHAR2','RAW','CHAR','NCHAR') then
            str:=str||' default '''||def||'''';
          elsif datatype='NUMBER' then
            str:=str||' default '||def;
          end if;
        end if;
        if tcol(x).nullable='N' then
          str:=str||' NOT NULL';
        end if;
      end if;
    end loop;
    if tdef(1).iot_type='IOT' then -- Index Organized Table (who invented this ??)
      for x in nvl(tcon.first,0)..nvl(tcon.last,0) loop
        if tcon.exists(x) then
          if tcon(x).table_name=tdef(1).table_name and tcon(x).constraint_type='P' then
            cn:=tcon(x).constraint_name;
            tcon.delete(x);
          end if;
        end if;
      end loop;
      second:=FALSE;
      for x in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
        if tconcol.exists(x) then
          if tconcol(x).table_name=tdef(1).table_name then
            if second then
              str:=str||','||tconcol(x).column_name;
            else
              str:=str||', constraint '||'IOT_'||to_char(sysdate,'YYYYMMDDHH24MISS')||'_PK'||' primary key ('||tconcol(x).column_name;
              second:=TRUE;
              tconcol.delete(x);
            end if;
          end if;
        end if;
      end loop;
      for x in nvl(tind.first,0)..nvl(tind.last,0) loop
        if tind.exists(x) then
          Debug('tn='||orig_tn||', table_name='||tind(x).table_name||', index_type='''||tind(x).index_type||'''');
          if tind(x).table_name=tdef(1).table_name and tind(x).index_type='IOT - TOP' then
            if tdef(1).tablespace_name is null then
              tdef(1).tablespace_name:=tind(x).tablespace_name;
            end if;
            if tdef(1).pct_free=0 then
              tdef(1).pct_free:=tind(x).pct_free;
            end if;
            if tdef(1).ini_trans=0 then
              tdef(1).ini_trans:=tind(x).ini_trans;
            end if;
            if tdef(1).max_trans=0 then
              tdef(1).max_trans:=tind(x).max_trans;
            end if;
            if tdef(1).initial_extent is null then
              tdef(1).initial_extent:=tind(x).initial_extent;
            end if;
            if tdef(1).next_extent is null then
              tdef(1).next_extent:=tind(x).next_extent;
            end if;
            if tdef(1).min_extents is null then
              tdef(1).min_extents:=tind(x).min_extents;
            end if;
            if tdef(1).max_extents is null then
              tdef(1).max_extents:=tind(x).max_extents;
            end if;
            if tdef(1).pct_increase is null then
              tdef(1).pct_increase:=tind(x).pct_increase;
            end if;
            if tdef(1).freelists is null then
              tdef(1).freelists:=tind(x).freelists;
            end if;
            if tdef(1).freelist_groups is null then
              tdef(1).freelist_groups:=tind(x).freelist_groups;
            end if;
            if tdef(1).pct_used!=0 then
              tind(x).pct_threshold:=tdef(1).pct_used;
            end if;
          end if;
        end if;
      end loop;
      str:=str||')';
    end if;
    Debug('CreateTable 1');
    str:=str||chr(10)||')'||chr(10);
    if tdef(1).iot_type='IOT' then -- Index Organized Table
      str:=str||'organization index';
      for x in nvl(tind.first,0)..nvl(tind.last,0) loop
        if tind.exists(x) then
          if tind(x).table_name=tdef(1).table_name and tind(x).index_type='IOT - TOP' then
            ipct:=tind(x).pct_threshold;
            for y in nvl(tcol.first,0)..nvl(tcol.last,0) loop
              if tcol.exists(y) then
                if tcol(y).column_id=tind(x).include_column then
                  ic:=tcol(y).column_name;
                  exit;
                end if;
              end if;
            end loop;
            for y in nvl(tindcol.first,0)..nvl(tindcol.last,0) loop
              if tindcol.exists(y) then
                if tindcol(y).index_name=tind(x).index_name then
                  tindcol.delete(y);
                end if;
              end if;
            end loop;
            tind.delete(x);
            exit;
          end if;
        end if;
      end loop;
      str:=str||CHR(10)||'pctthreshold '||to_char(ipct);
      if ic is not null then
        str:=str||CHR(10)||'including '||ic;
      end if;
      select min(tablespace_name) into tsn from sys.all_tables
      where iot_name=orig_tn
      and owner=tdef(1).owner;
      if tsn is not null then
        str:=str||chr(10)||'overflow';
      end if;
      str:=str||CHR(10);
    end if;
    str:=str||'tablespace '||tdef(1).tablespace_name||chr(10);
    str:=str||'pctfree '||to_char(tdef(1).pct_free)||chr(10);
    if tdef(1).iot_type!='IOT' then
      str:=str||'pctused '||to_char(tdef(1).pct_used)||chr(10);
    end if;
    str:=str||'initrans '||to_char(tdef(1).ini_trans)||chr(10);
    str:=str||'maxtrans '||to_char(tdef(1).max_trans)||chr(10);
    str:=str||'storage (initial '||to_char(tdef(1).initial_extent)||chr(10);
    str:=str||'         next '||to_char(tdef(1).next_extent)||chr(10);
    str:=str||'         minextents '||to_char(tdef(1).min_extents)||chr(10);
    str:=str||'         maxextents '||to_char(tdef(1).max_extents)||chr(10);
    str:=str||'         pctincrease '||to_char(tdef(1).pct_increase)||chr(10);
    str:=str||'         freelists '||to_char(tdef(1).freelists)||chr(10);
    str:=str||'         freelist groups '||to_char(tdef(1).freelist_groups)||chr(10);
    str:=str||'        )';
    DDL(str);
    Debug('End CreateTable');
  END;  
  
  PROCEDURE CreateIndexes(idx_name VARCHAR2,
                          tind     TableIndexes,
                          tindcol  TableIndexColumns) IS
  str    VARCHAR2(32000);
  second BOOLEAN;
  BEGIN
    Debug('Start CreateIndexes');
    for x in nvl(tind.first,0)..nvl(tind.last,0) loop
      if tind.exists(x) then
        if idx_name is null or idx_name=tind(x).index_name then
          WorkingLog('Creating index '||tind(x).index_name);
          str:='create';
          if tind(x).index_type='BITMAP' then
            str:=str||' bitmap';
          elsif tind(x).uniqueness='UNIQUE' then
            str:=str||' unique';
          end if;
          str:=str||' index "';
          str:=str||tind(x).owner||'".'||tind(x).index_name||chr(10)||'on "';
          str:=str||tind(x).table_owner||'".'||tind(x).table_name||' ('||chr(10)||' ';
          second:=false;
          for y in nvl(tindcol.first,0)..nvl(tindcol.last,0) loop
            if tindcol.exists(y) and tind(x).index_name=tindcol(y).index_name then
              if second then
                str:=str||','||chr(10)||' ';
              else
                second:=true;
              end if;
              str:=str||tindcol(y).column_name;
            end if;
          end loop;
          if not second then
            return;
          end if;
          str:=str||')';
          if tind(x).tablespace_name is not null then
            str:=str||chr(10)||'tablespace '||tind(x).tablespace_name;
          end if;
          if tind(x).ini_trans is not null then
            str:=str||chr(10)||'initrans '||to_char(tind(x).ini_trans);
          end if;
          if tind(x).max_trans is not null then
            str:=str||chr(10)||'maxtrans '||to_char(tind(x).max_trans);
          end if;
          if tind(x).initial_extent is not null or
             tind(x).next_extent is not null or
             tind(x).min_extents is not null or
             tind(x).max_extents is not null or
             tind(x).pct_increase is not null or
             tind(x).freelists is not null or
             tind(x).freelist_groups is not null then
            str:=str||chr(10)||'storage ('||chr(10);
            if tind(x).initial_extent is not null then
              str:=str||'  initial '||to_char(tind(x).initial_extent)||chr(10);
            end if;
            if tind(x).next_extent is not null then
              str:=str||'  next '||to_char(tind(x).next_extent)||chr(10);
            end if;
            if tind(x).min_extents is not null then
              str:=str||'  minextents '||to_char(tind(x).min_extents)||chr(10);
            end if;
            if tind(x).max_extents is not null then
              str:=str||'  maxextents '||to_char(tind(x).max_extents)||chr(10);
            end if;
            if tind(x).pct_increase is not null then
              str:=str||'  pctincrease '||to_char(tind(x).pct_increase)||chr(10);
            end if;
            if tind(x).freelists is not null then
              str:=str||'  freelists '||to_char(tind(x).freelists)||chr(10);
            end if;
            if tind(x).freelist_groups is not null then
              str:=str||'  freelist groups '||to_char(tind(x).freelist_groups)||chr(10);
            end if;
            str:=str||')';
          end if;
          if tind(x).index_type in ('NORMAL/REV','BITMAP/REV') then
            str:=str||chr(10)||' reverse';
          end if;
          BEGIN
            DDL(str);
          EXCEPTION WHEN OTHERS THEN
            AddToLog('Creation of index '||tind(x).index_name||': '||sqlerrm);
          END;
        end if;
      end if;
    end loop;
    Debug('End CreateIndexes');
  END;

  PROCEDURE CreateConstraints(tcon    TableConstraints,
                              tconcol TableConstraintColumns) IS
  str      VARCHAR2(32000);
  second   BOOLEAN;
  ctype    VARCHAR2(1);
  cn       VARCHAR2(30);
  tn       VARCHAR2(30);
  owner_id VARCHAR2(30);
  cursor c(tn VARCHAR2,cn VARCHAR2) is 
    select column_name from sys.all_cons_columns
    where table_name=tn and constraint_name=cn order by position;
  BEGIN
    Debug('Start CreateConstraints');
    for x in nvl(tcon.first,0)..nvl(tcon.last,0) loop
      if tcon.exists(x) then
        WorkingLog('Creating constraint '||upper(tcon(x).constraint_name));
        str:='alter table "'||tcon(x).owner||'".'||tcon(x).table_name||' add constraint ';
        str:=str||tcon(x).constraint_name;
        ctype:=tcon(x).constraint_type;
        Debug('Constraint '||tcon(x).constraint_name||', type='||ctype);
        if ctype='C' then
          str:=str||' CHECK (';
          str:=str||tcon(x).search_condition;
          str:=str||')';
          str:=str||' '||tcon(x).deferrable||' ';
          str:=str||' INITIALLY '||tcon(x).deferred;
          if tcon(x).status='ENABLED' then
            str:=str||' ENABLE';
            if tcon(x).validated='VALIDATE' then
              str:=str||' VALIDATE';
            else
              str:=str||' NOVALIDATE';
            end if;
          else
            str:=str||' DISABLE';
          end if;
          DDL(str);
        elsif ctype='P' then
          str:=str||' PRIMARY KEY (';
          second:=false;
          for y in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
            if tconcol.exists(y) and tcon(x).constraint_name=tconcol(y).constraint_name then
              if second then
                str:=str||',';
              else
                second:=true;
              end if;
              str:=str||tconcol(y).column_name;
            end if;
          end loop;
          str:=str||')';
          str:=str||' '||tcon(x).deferrable||' ';
          str:=str||' INITIALLY '||tcon(x).deferred;
          if tcon(x).status='ENABLED' then
            str:=str||' ENABLE';
            if tcon(x).validated='VALIDATE' then
              str:=str||' VALIDATE';
            else
              str:=str||' NOVALIDATE';
            end if;
          else
            str:=str||' DISABLE';
          end if;
          DDL(str);
        elsif ctype='U' then
          str:=str||' UNIQUE (';
          second:=false;
          for y in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
            if tconcol.exists(y) and tcon(x).constraint_name=tconcol(y).constraint_name then
              if second then
                str:=str||',';
              else
                second:=true;
              end if;
              str:=str||tconcol(y).column_name;
            end if;
          end loop;
          str:=str||')';
          str:=str||' '||tcon(x).deferrable||' ';
          str:=str||' INITIALLY '||tcon(x).deferred;
          if tcon(x).status='ENABLED' then
            str:=str||' ENABLE';
            if tcon(x).validated='VALIDATE' then
              str:=str||' VALIDATE';
            else
              str:=str||' NOVALIDATE';
            end if;
          else
            str:=str||' DISABLE';
          end if;
          DDL(str);
        elsif ctype='R' then
          str:=str||' FOREIGN KEY (';
          second:=false;
          for y in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
            if tconcol.exists(y) and tcon(x).constraint_name=tconcol(y).constraint_name then
              if second then
                str:=str||',';
              else
                second:=true;
              end if;
              str:=str||tconcol(y).column_name;
            end if;
          end loop;
          str:=str||') REFERENCES ';
          cn:=tcon(x).r_constraint_name;
          begin
            select owner,table_name into owner_id,tn from sys.all_constraints where constraint_name=cn;
            str:=str||'"'||owner_id||'".'||tn||' (';
            second:=false;
            for z in c(tn,cn) loop
              if second then
                str:=str||',';
              else
                second:=true;
              end if;
              str:=str||z.column_name;
            end loop;
            str:=str||')';
            if tcon(x).delete_rule='CASCADE' then
              str:=str||' ON DELETE CASCADE ';
            end if;
            str:=str||' '||tcon(x).deferrable||' ';
            str:=str||' INITIALLY '||tcon(x).deferred;
            if tcon(x).status='ENABLED' then
              str:=str||' ENABLE';
              if tcon(x).validated='VALIDATE' then
                str:=str||' VALIDATE';
              else
                str:=str||' NOVALIDATE';
              end if;
            else
              str:=str||' DISABLE';
            end if;
          exception when others then --Circular reference
            null;
          end;
        end if;
      end if;
    end loop;
    Debug('End CreateConstraints');
  END;

  PROCEDURE CreateForeignConstraints(table_name VARCHAR2,
                                     tfcon      TableFConstraints,
                                     tfconcol   TableFConstraintColumns,
                                     tcon       TableConstraints,
                                     tconcol    TableConstraintColumns) IS
  str VARCHAR2(32000);
  second BOOLEAN;
  ctype VARCHAR2(1);
  cn VARCHAR2(30);
  tn VARCHAR2(30);
  BEGIN
    Debug('Start CreateForeignConstraints');
    for x in nvl(tfcon.first,0)..nvl(tfcon.last,0) loop
      if tfcon.exists(x) then
        if tfcon(x).table_name!=table_name then
          WorkingLog('Creating foreign constraint '||tfcon(x).constraint_name);
          str:='alter table "'||tfcon(x).owner||'".'||tfcon(x).table_name||' add constraint ';
          str:=str||tfcon(x).constraint_name;
          str:=str||' FOREIGN KEY (';
          second:=false;
          for y in nvl(tfconcol.first,0)..nvl(tfconcol.last,0) loop
            if tfconcol.exists(y) and tfcon(x).constraint_name=tfconcol(y).constraint_name then
              if second then
                str:=str||',';
              else
                second:=true;
              end if;
              str:=str||tfconcol(y).column_name;
            end if;
          end loop;
          str:=str||') REFERENCES ';
          cn:=tfcon(x).r_constraint_name;
          for y in nvl(tcon.first,0)..nvl(tcon.last,0) loop
            if tcon.exists(y) and tcon(y).constraint_name=cn then
              str:=str||'"'||tcon(y).owner||'".'||tcon(y).table_name;
              exit;
            end if;
          end loop;
          str:=str||' (';
          second:=false;
          for y in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
            if tconcol.exists(y) and tconcol(y).constraint_name=cn then
              if second then
                str:=str||',';
              else
                second:=true;
              end if;
              str:=str||tconcol(y).column_name;
            end if;
          end loop;
          str:=str||')';
          if tfcon(x).delete_rule='CASCADE' then
            str:=str||' ON DELETE CASCADE ';
          end if;
          str:=str||' '||tfcon(x).deferrable||' ';
          str:=str||' INITIALLY '||tfcon(x).deferred;
          if tfcon(x).status='ENABLED' then
            str:=str||' ENABLE';
            if tfcon(x).validated='VALIDATE' then
              str:=str||' VALIDATE';
            else
              str:=str||' NOVALIDATE';
            end if;
          else
              str:=str||' DISABLE';
          end if;
          DDL(str);
        end if;
      end if;
    end loop;
    Debug('Start CreateForeignConstraints');
  END;

  PROCEDURE CreateComments(tcom TableComments,tcolcom TableColumnComments) IS
  str VARCHAR2(32000);
  BEGIN
    Debug('Start CreateComments');
    WorkingLog('Creating comments');
    if tcom.count>0 then
      str:='comment on table "'||tcom(1).owner||'".'||tcom(1).table_name||' is ''';
      str:=str||tcom(1).comments;
      str:=str||'''';
      DDL(str);
    end if;
    for x in nvl(tcolcom.first,0)..nvl(tcolcom.last,0) loop
      if tcolcom.exists(x) then
        str:='comment on column "'||tcolcom(x).owner||'".'||tcolcom(x).table_name||'.';
        str:=str||tcolcom(x).column_name;
        str:=str||' is ''';
        str:=str||tcolcom(x).comments;
        str:=str||'''';
        DDL(str);
      end if;
    end loop;
    Debug('End CreateComments');
  END;

  PROCEDURE CreateTriggers(ttrig TableTriggers) IS
  str VARCHAR2(32000);
  wc VARCHAR2(4000);
  BEGIN
    Debug('Start CreateTriggers');
    for x in nvl(ttrig.first,0)..nvl(ttrig.last,0) loop
      if ttrig.exists(x) then
        WorkingLog('Creating trigger '||ttrig(x).owner||'.'||ttrig(x).trigger_name);
        str:='create or replace trigger';
        str:=str||' "'||ttrig(x).owner||'".'||replace(upper(ttrig(x).description),'ON '||ttrig(x).table_name,' ON "'||ttrig(x).table_owner||'".'||ttrig(x).table_name);
        wc:=ttrig(x).when_clause;
        if wc is not null then
          wc:=substr(wc,1,length(wc)-1);
          str:=str||' when ('||wc;
          str:=str||')';
        end if;
        str:=str||' '||ttrig(x).trigger_body;
        BEGIN
          DDL(str);
        EXCEPTION WHEN OTHERS THEN
          AddToLog('Creation of trigger '||ttrig(x).trigger_name||': '||sqlerrm);
        END;
        if ttrig(x).status='DISABLED' then
          str:='alter trigger '||ttrig(x).trigger_name||' disable';
        BEGIN
          DDL(str);
        EXCEPTION WHEN OTHERS THEN
          AddToLog('Disabeling of trigger '||ttrig(x).trigger_name||': '||sqlerrm);
        END;
        end if;
      end if;
    end loop;
    Debug('End CreateTriggers');
  END;

  PROCEDURE CreateSynonyms(tsyn TableSynonyms) IS
  str VARCHAR2(32000);
  BEGIN
    Debug('Start CreateSynonyms');
    WorkingLog('Creating synonyms');
    for x in nvl(tsyn.first,0)..nvl(tsyn.last,0) loop
      if tsyn.exists(x) then
        if tsyn(x).owner='PUBLIC' then
          str:='create public synonym '||tsyn(x).synonym_name||' for "'||tsyn(x).table_owner||'".'||tsyn(x).table_name;
        else
          str:='create synonym "'||tsyn(x).owner||'".'||tsyn(x).synonym_name||' for "'||tsyn(x).table_owner||'".'||tsyn(x).table_name;
        end if;
        begin
          DDL(str);
        exception when others then
          AddToLog('Creation of synonym '||tsyn(x).synonym_name||': '||sqlerrm);
        end;
      end if;
    end loop;
    Debug('End CreateSynonyms');
  END;

  PROCEDURE DropSynonyms(tsyn TableSynonyms) IS
  str VARCHAR2(32000);
  BEGIN
    Debug('Start DropSynonyms');
    WorkingLog('Dropping synonyms');
    for x in nvl(tsyn.first,0)..nvl(tsyn.last,0) loop
      if tsyn.exists(x) then
        if tsyn(x).owner='PUBLIC' then
          str:='drop public synonym '||tsyn(x).synonym_name;
        else
          str:='drop synonym "'||tsyn(x).owner||'".'||tsyn(x).synonym_name;
        end if;
        begin
          DDL(str);
        exception when others then
          AddToLog('Drop of synonym '||tsyn(x).synonym_name||': '||sqlerrm);
        end;
      end if;
    end loop;
    Debug('End DropSynonyms');
  END;

  PROCEDURE CreateGrants(tgrnt TableGrants) IS
  str VARCHAR2(32000);
  BEGIN
    Debug('Start CreateGrants');
    WorkingLog('Granting priviliges');
    for x in nvl(tgrnt.first,0)..nvl(tgrnt.last,0) loop
      if tgrnt.exists(x) then
        str:='grant '||tgrnt(x).privilege||' on "'||tgrnt(x).table_schema||'".'||tgrnt(x).table_name||' to "'||tgrnt(x).grantee||'"';
        if tgrnt(x).grantable='YES' then
          str:=str||' with grant option';
        end if;
        begin
          DDL(str);
        exception when others then
          AddToLog('granting '||tgrnt(x).privilege||' to '||tgrnt(x).grantee||': '||sqlerrm);
        end;
      end if;
    end loop;
    Debug('End CreateGrants');
  END;

  PROCEDURE CreateColumnGrants(tcgrnt TableColumnGrants) IS
  str VARCHAR2(32000);
  BEGIN
    Debug('Start CreateColumnGrants');
    WorkingLog('Granting column priviliges');
    for x in nvl(tcgrnt.first,0)..nvl(tcgrnt.last,0) loop
      if tcgrnt.exists(x) then
        str:='grant '||tcgrnt(x).privilege||' ('||tcgrnt(x).column_name||') on "'||tcgrnt(x).table_schema||'".'||tcgrnt(x).table_name||' to "'||tcgrnt(x).grantee||'"';
        if tcgrnt(x).grantable='YES' then
          str:=str||' with grant option';
        end if;
        begin
          DDL(str);
        exception when others then
          AddToLog('granting '||tcgrnt(x).privilege||' to '||tcgrnt(x).grantee||': '||sqlerrm);
        end;
      end if;
    end loop;
    Debug('End CreateColumnGrants');
  END;

  PROCEDURE AlterTableName(old_tn          VARCHAR2,
                           table_name      VARCHAR2,
                           tcol     IN OUT TableColumns,
                           tind     IN OUT TableIndexes,
                           tindcol  IN OUT TableIndexColumns,
                           tcon     IN OUT TableConstraints,
                           tconcol  IN OUT TableConstraintColumns,
                           ttrig    IN OUT TableTriggers,
                           ttrigcol IN OUT TableTriggerColumns,
                           tcolcom  IN OUT TableColumnComments,
                           tcom     IN OUT TableComments,
                           tfcon    IN OUT TableFConstraints,
                           tfconcol IN OUT TableFConstraintColumns,
                           tsyn     IN OUT TableSynonyms,
                           tgrnt    IN OUT TableGrants,
                           tcgrnt   IN OUT TableColumnGrants) IS
  BEGIN
    for x in nvl(tcol.first,0)..nvl(tcol.last,0) loop
      if tcol.exists(x) then
        tcol(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tind.first,0)..nvl(tind.last,0) loop
      if tind.exists(x) then
        tind(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tindcol.first,0)..nvl(tindcol.last,0) loop
      if tindcol.exists(x) then
        tindcol(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tcon.first,0)..nvl(tcon.last,0) loop
      if tcon.exists(x) then
        tcon(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tconcol.first,0)..nvl(tconcol.last,0) loop
      if tconcol.exists(x) then
        tconcol(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(ttrig.first,0)..nvl(ttrig.last,0) loop
      if ttrig.exists(x) then
        ttrig(x).table_name:=table_name;
        ttrig(x).when_clause:=replace(upper(ttrig(x).when_clause),upper(old_tn),upper(table_name));
        ttrig(x).description:=replace(upper(ttrig(x).description),upper(old_tn),upper(table_name));
        ttrig(x).trigger_body:=replace(upper(ttrig(x).trigger_body),upper(old_tn),upper(table_name));
      end if;
    end loop;
    for x in nvl(ttrigcol.first,0)..nvl(ttrigcol.last,0) loop
      if ttrigcol.exists(x) then
        ttrigcol(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tcolcom.first,0)..nvl(tcolcom.last,0) loop
      if tcolcom.exists(x) then
        tcolcom(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tcom.first,0)..nvl(tcom.last,0) loop
      if tcom.exists(x) then
        tcom(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tfcon.first,0)..nvl(tfcon.last,0) loop
      if tfcon.exists(x) then
        tfcon(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tfconcol.first,0)..nvl(tfconcol.last,0) loop
      if tfconcol.exists(x) then
        tfconcol(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tsyn.first,0)..nvl(tsyn.last,0) loop
      if tsyn.exists(x) then
        tsyn(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tgrnt.first,0)..nvl(tgrnt.last,0) loop
      if tgrnt.exists(x) then
        tgrnt(x).table_name:=table_name;
      end if;
    end loop;
    for x in nvl(tcgrnt.first,0)..nvl(tcgrnt.last,0) loop
      if tcgrnt.exists(x) then
        tcgrnt(x).table_name:=table_name;
      end if;
    end loop;
  end;
  
  PROCEDURE RecreateTable(table_name      VARCHAR2,
                          notcolumn       VARCHAR2,
                          nottcolumn      VARCHAR2,
                          docolumn        VARCHAR2,
                          conv_function   VARCHAR2,
                          tdef     IN OUT TableDef,
                          tcol     IN OUT TableColumns,
                          tind     IN OUT TableIndexes,
                          tindcol  IN OUT TableIndexColumns,
                          tcon     IN OUT TableConstraints,
                          tconcol  IN OUT TableConstraintColumns,
                          ttrig    IN OUT TableTriggers,
                          ttrigcol IN OUT TableTriggerColumns,
                          tcolcom  IN OUT TableColumnComments,
                          tcom     IN OUT TableComments,
                          tfcon    IN OUT TableFConstraints,
                          tfconcol IN OUT TableFConstraintColumns,
                          tsyn     IN OUT TableSynonyms,
                          tgrnt    IN OUT TableGrants,
                          tcgrnt   IN OUT TableColumnGrants) IS
  cnt1      NUMBER;
  cnt2      NUMBER;
  tmp       NUMBER;
  o_tind    TableIndexes;
  o_tindcol TableIndexColumns;
  o_tcon    TableConstraints;
  o_tconcol TableConstraintColumns;
  BEGIN
    Debug('Start RecreateTable');
    o_tind:=tind;
    o_tindcol:=tindcol;
    o_tcon:=tcon;
    o_tconcol:=tconcol;
    if upper(table_name)!=tdef(1).table_name then
      AlterTableName(table_name,tdef(1).table_name,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol,tsyn,tgrnt,tcgrnt);
    end if;
    DropTemporaryTable;
    CreateTable(temporary_table,tdef,tcol,temporary_table_owner,tind,tindcol,tcon,tconcol,upper(table_name));
    cnt1:=CountTable(table_name,NULL,tdef(1).owner);
    CopyTable(table_name,temporary_table,notcolumn,nottcolumn,TRUE,docolumn,conv_function,tcol,tdef(1).owner,temporary_table_owner);
    DDL('commit');
    cnt2:=CountTable(temporary_table,NULL,tdef(1).owner);
    if cnt1!=cnt2 then
      DropTemporaryTable;
      ApplicationError(-20003,to_char(cnt2-cnt1)||' records would be lost, Operation aborted');
      return;
    end if;
    DDL('drop table "'||tdef(1).owner||'".'||table_name||' cascade constraints');  
    begin
      DDL('rename '||temporary_table||' to '||tdef(1).table_name);  
    exception when others then
      WorkingLog('Creating table '||tdef(1).table_name);
      tind:=o_tind;
      tindcol:=o_tindcol;
      tcon:=o_tcon;
      tconcol:=o_tconcol;
      CreateTable(tdef(1).table_name,tdef,tcol,tdef(1).owner,tind,tindcol,tcon,tconcol,upper(temporary_table));
      WorkingLog('Coping data back to table '||tdef(1).table_name);
      CopyTable(temporary_table,tdef(1).table_name,NULL,NULL,FALSE,NULL,NULL,tcol,tdef(1).owner,tdef(1).owner);
      DDL('commit');
    end;
    DropSynonyms(tsyn);
    CreateIndexes(NULL,tind,tindcol);
    CreateConstraints(tcon,tconcol);
    CreateForeignConstraints(table_name,tfcon,tfconcol,tcon,tconcol);
    CreateComments(tcom,tcolcom);
    CreateTriggers(ttrig);
    CreateSynonyms(tsyn);
    CreateGrants(tgrnt);
    CreateColumnGrants(tcgrnt);
    Debug('End RecreateTable');
  END;

  FUNCTION DropConstraint_I(cn           IN      VARCHAR2,
  			    owner_id     IN      VARCHAR2,
                            keep_record  IN      BOOLEAN DEFAULT FALSE,
                            del_cons     IN OUT TableConstraints,
                            del_cons_col IN OUT TableConstraintColumns) RETURN BOOLEAN IS
  tn       VARCHAR2(30);
  cursor c(cn VARCHAR2) is select constraint_name,owner
                           from all_constraints
                           where r_constraint_name=upper(cn);
  cursor d(cn VARCHAR2) is select *
                           from all_cons_columns
                           where constraint_name=upper(cn);
  BEGIN
    Debug('Start DropConstraint_I');
    
    BEGIN
      select table_name into tn 
      from all_constraints 
      where constraint_name=upper(cn)
      and owner=owner_id;
    EXCEPTION WHEN OTHERS THEN
      Debug('Looking up table name: '||sqlerrm);
      ApplicationError(-20007,'Constraint '||upper(cn)||' does not exist');
      return FALSE;
    END;

    for x in c(cn) loop
      Debug('Dropping foreign constraint '||x.constraint_name||' that references this constraint');
      if not DropConstraint_I(x.constraint_name,x.owner,keep_record,del_cons,del_cons_col) then
        return FALSE;
      end if;
    end loop;
    if keep_record then
      Debug('keeping record ('||to_char(nvl(del_cons.last,0)+1)||') of constraint');
      select * into del_cons(nvl(del_cons.last,0)+1) from sys.all_constraints where constraint_name=upper(cn);
      for x in d(cn) loop
        del_cons_col(nvl(del_cons_col.last,0)+1):=x;
      end loop;
    end if;
    WorkingLog('Dropping constraint '||upper(cn)||' on table '||tn);
    DDL('alter table "'||owner_id||'".'||tn||' drop constraint '||cn);
    Debug('End DropConstraint_I');
    return TRUE;
  END;

  FUNCTION DropIndex_I(ind          IN     VARCHAR2,
                       keep_record  IN     BOOLEAN DEFAULT FALSE,
                       del_cons     IN OUT TableConstraints,
                       del_cons_col IN OUT TableConstraintColumns,
                       owner_id     IN     VARCHAR2,
                       index_type   IN     VARCHAR2 DEFAULT 'NORMAL') RETURN BOOLEAN IS
  cn VARCHAR2(30);
  tn VARCHAR2(30);
  tr VARCHAR2(30);
  ct VARCHAR2(1);
  BEGIN
    Debug('Start DropIndex_I');
    Debug('Checking for constraint index');
    BEGIN
      select table_name,table_owner into tn,tr from all_indexes where index_name=upper(ind) and owner=upper(owner_id);

      select distinct uc.constraint_name,uc.constraint_type
      into cn,ct
      from all_constraints uc,
           (select count(1) cnt from all_ind_columns
            where index_name=upper(ind)
            and index_owner=upper(owner_id)
            and table_name=tn
            and table_owner=upper(owner_id)) ic
      where uc.table_name=tn
      and uc.owner=upper(owner_id)
      and uc.constraint_type in ('P','U')
      and ic.cnt=(select count(1) from all_cons_columns ucc,all_ind_columns uic
                  where ucc.constraint_name=uc.constraint_name
                  and uic.column_name=ucc.column_name
                  and uic.index_owner=upper(owner_id)
                  and ucc.owner=upper(owner_id)
                  and uic.index_name=upper(ind));
       
      if ct='P' and index_type='BITMAP' then
        WorkingLog('Bitmap indexes can''t be used for primary keys');
        return FALSE;
      else
        WorkingLog('Index '||upper(ind)||' is needed for constraint enforcement, dropping constraint instead');
        if not DropConstraint_I(cn,owner_id,keep_record,del_cons,del_cons_col) then
          return FALSE;
        end if;
      end if;
    EXCEPTION WHEN NO_DATA_FOUND THEN
      WorkingLog('Dropping index '||upper(ind));
      DDL('drop index '||ind);
    END;
    Debug('End DropIndex_I');
    return TRUE;
  END;

  FUNCTION DropTrigger_I(trg      IN VARCHAR2,
                         owner_id IN VARCHAR2) RETURN BOOLEAN IS
  BEGIN
    Debug('Start DropTrigger_I');
    WorkingLog('Dropping trigger '||upper(trg));
    DDL('drop trigger "'||upper(owner_id)||'".'||trg);
    Debug('End DropTrigger_I');
    return TRUE;
  END;

  FUNCTION DropColumnComment_I(tn       IN VARCHAR2,
                               cn       IN VARCHAR2,
                               owner_id IN VARCHAR2) RETURN BOOLEAN IS
  BEGIN
    Debug('Start DropColumnComment_I');
    WorkingLog('Dropping comment on column '||upper(tn)||'.'||upper(cn));
    DDL('comment on column "'||owner_id||'".'||tn||'.'||cn||' is ''''');
    Debug('End DropColumnComment_I');
    return TRUE;
  END;

  FUNCTION DropTableComment_I(tn       IN VARCHAR2,
                              owner_id IN VARCHAR2) RETURN BOOLEAN IS
  BEGIN
    Debug('Start DropTableComment_I');
    WorkingLog('Dropping comment on table '||upper(tn));
    DDL('comment on table "'||owner_id||'".'||tn||' is ''''');
    Debug('End DropTableComment_I');
    return TRUE;
  END;

  FUNCTION AlterTable_I(tn IN VARCHAR2,
                        o TableDef,
                        n IN OUT TableDef) RETURN BOOLEAN IS
  tdef     TableDef;
  tcol     TableColumns;
  tind     TableIndexes;
  tindcol  TableIndexColumns;
  tcon     TableConstraints;
  tconcol  TableConstraintColumns;
  ttrig    TableTriggers;
  ttrigcol TableTriggerColumns;
  tcolcom  TableColumnComments;
  tcom     TableComments;
  tfcon    TableFConstraints;
  tfconcol TableFConstraintColumns;
  tsyn     TableSynonyms;
  tgrnt    TableGrants;
  tcgrnt   TableColumnGrants;
  str      VARCHAR2(4000);
  BEGIN
    Debug('Start AlterTable_I');
    if  o(1).tablespace_name!=n(1).tablespace_name
     or o(1).initial_extent!=n(1).initial_extent
     or o(1).owner!=n(1).owner
     or o(1).freelists!=n(1).freelists
     or o(1).freelist_groups!=n(1).freelist_groups
     or o(1).min_extents<n(1).min_extents
     or o(1).freelist_groups!=n(1).freelist_groups
     or n(1).iot_type='IOT' then
      GetTableDefinition(tn,tdef,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol,tsyn,tgrnt,tcgrnt,o(1).owner);
      RecreateTable(tn,null,null,null,null,n,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol,tsyn,tgrnt,tcgrnt);
    else
      if o(1).table_name!=n(1).table_name then
        WorkingLog('Renaming table '||o(1).table_name||' to '||n(1).table_name);
        begin
          GetTableSynonyms(tn,tsyn,o(1).owner);
          DDL('rename "'||o(1).owner||'".'||o(1).table_name||' to "'||n(1).owner||'".'||n(1).table_name);
          DropSynonyms(tsyn);
          CreateSynonyms(tsyn);
        exception when others then
          GetTableDefinition(tn,tdef,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol,tsyn,tgrnt,tcgrnt,o(1).owner);
          RecreateTable(tn,null,null,null,null,n,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol,tsyn,tgrnt,tcgrnt);
        end;
      end if;
      str:='alter table "'||o(1).owner||'".'||tn;
      if o(1).pct_free!=n(1).pct_free then
        str:=str||' pctfree '||to_char(n(1).pct_free);
      end if;
      if o(1).pct_used!=n(1).pct_used then
        str:=str||' pctused '||to_char(n(1).pct_used);
      end if;
      if o(1).ini_trans!=n(1).ini_trans then
        str:=str||' initrans '||to_char(n(1).ini_trans);
      end if;
      if o(1).max_trans!=n(1).max_trans then
        str:=str||' maxtrans '||to_char(n(1).max_trans);
      end if;
      if  o(1).next_extent!=n(1).next_extent
       or o(1).min_extents!=n(1).min_extents
       or o(1).pct_increase!=n(1).pct_increase
       or o(1).max_extents!=n(1).max_extents then
        str:=str||' storage (';
        if o(1).next_extent!=n(1).next_extent then
          str:=str||' next '||to_char(n(1).next_extent);
        end if;
        if o(1).min_extents!=n(1).min_extents then
          str:=str||' minextents '||to_char(n(1).min_extents);
        end if;
        if o(1).max_extents!=n(1).max_extents then
          str:=str||' maxextents '||to_char(n(1).max_extents);
        end if;
        if o(1).pct_increase!=n(1).pct_increase then
          str:=str||' pctincrease '||to_char(n(1).pct_increase);
        end if;
        str:=str||')';
      end if;
      if str!='alter table "'||o(1).owner||'".'||tn then
        WorkingLog('Altering table '||upper(tn));
        DDL(str);
      end if;
    end if;
    Debug('End AlterTable_I');
    return TRUE;
  END;

  FUNCTION AlterIndex_I(ind IN VARCHAR2,o TableIndexes,n TableIndexes,oc TableIndexColumns,nc TableIndexColumns) RETURN BOOLEAN IS
  str      VARCHAR2(4000);
  tc TableConstraints;
  tcc TableConstraintColumns;
  same_columns BOOLEAN:=TRUE;
  BEGIN
    Debug('Start AlterIndex_I');
    for x in nvl(oc.first,0)..nvl(oc.last,0) loop
      if oc.exists(x) then
        if not nc.exists(x) then
          same_columns:=FALSE;
          exit;
        end if;
        if  oc(x).column_name!=nc(x).column_name
         or oc(x).column_position!=nc(x).column_position then
          same_columns:=FALSE;
          exit;
        end if;
      elsif nc.exists(x) then
        same_columns:=FALSE;
        exit;
      end if;
    end loop;
    if  o(1).tablespace_name!=n(1).tablespace_name
     or o(1).initial_extent!=n(1).initial_extent
     or o(1).owner!=n(1).owner
     or o(1).freelists!=n(1).freelists
     or o(1).freelist_groups!=n(1).freelist_groups
     or o(1).min_extents<n(1).min_extents
     or o(1).uniqueness!=n(1).uniqueness
     or o(1).index_type!=n(1).index_type
     or o(1).index_name!=n(1).index_name
     or not same_columns
     or o(1).freelist_groups!=n(1).freelist_groups then
      if not DropIndex_I(ind,TRUE,tc,tcc,o(1).owner,n(1).index_type) then
        return FALSE;
      end if;
      CreateIndexes(ind,n,nc);
      CreateConstraints(tc,tcc);
    else
      str:='alter index "'||n(1).owner||'".'||ind;
      if o(1).pct_free!=n(1).pct_free then
        str:=str||' pctfree '||to_char(n(1).pct_free);
      end if;
      if o(1).ini_trans!=n(1).ini_trans then
        str:=str||' initrans '||to_char(n(1).ini_trans);
      end if;
      if o(1).max_trans!=n(1).max_trans then
        str:=str||' maxtrans '||to_char(n(1).max_trans);
      end if;
      if  o(1).next_extent!=n(1).next_extent
       or o(1).min_extents!=n(1).min_extents
       or o(1).pct_increase!=n(1).pct_increase
       or o(1).max_extents!=n(1).max_extents then
        str:=str||' storage (';
        if o(1).next_extent!=n(1).next_extent then
          str:=str||' next '||to_char(n(1).next_extent);
        end if;
        if o(1).min_extents!=n(1).min_extents then
          str:=str||' minextents '||to_char(n(1).min_extents);
        end if;
        if o(1).max_extents!=n(1).max_extents then
          str:=str||' maxextents '||to_char(n(1).max_extents);
        end if;
        if o(1).pct_increase!=n(1).pct_increase then
          str:=str||' pctincrease '||to_char(n(1).pct_increase);
        end if;
        str:=str||')';
      end if;
      if str!='alter index "'||n(1).owner||'".'||ind then
        WorkingLog('Altering index '||upper(ind));
        DDL(str);
      end if;
    end if;
    Debug('End AlterIndex_I');
    return TRUE;
  END;

  PROCEDURE Grep_I(table_name VARCHAR2,owner VARCHAR2,column_name VARCHAR2,what VARCHAR2,casesens BOOLEAN,result VARCHAR2,datatype VARCHAR2) IS
  str VARCHAR2(32000);
  BEGIN
    str:='INSERT INTO '||result||' (owner,table_name,column_name,row_id,text) SELECT ';
    if datatype IN ('VARCHAR2','CHAR') then
      str:=str||''''||owner||''','''||table_name||''','''||column_name||''',rowid,'||column_name;
    elsif datatype='NUMBER' then
      str:=str||''''||owner||''','''||table_name||''','''||column_name||''',rowid,to_char('||column_name||')';
    elsif datatype='DATE' then
      str:=str||''''||owner||''','''||table_name||''','''||column_name||''',rowid,to_char('||column_name||',''dd/mm/yyyy hh24:mi:ss'')';
    end if;
    str:=str||' FROM "'||owner||'".'||table_name||' WHERE ';
    if casesens then
      if datatype in ('VARCHAR2','CHAR') then
        str:=str||column_name||' LIKE ''%'||what||'%''';
      elsif datatype='NUMBER' then
        str:=str||'to_char('||column_name||') LIKE ''%'||what||'%''';
      elsif datatype='DATE' then
        str:=str||'to_char('||column_name||',''dd/mm/yyyy hh24:mi:ss'') LIKE ''%'||what||'%''';
      end if;
    else
      if datatype in ('VARCHAR2','CHAR') then
        str:=str||'upper('||column_name||') LIKE ''%'||upper(what)||'%''';
      elsif datatype='NUMBER' then
        str:=str||'upper(to_char('||column_name||')) LIKE ''%'||upper(what)||'%''';
      elsif datatype='DATE' then
        str:=str||'upper(to_char('||column_name||',''dd/mm/yyyy hh24:mi:ss'')) LIKE ''%'||upper(what)||'%''';
      end if;
    end if;
    DDL(str);
  END;

--------------------------------------------------
--                                              --
-- FROM THIS POINT COME THE EXPORTED PROCEDURES --
--                                              --
--------------------------------------------------
 
  FUNCTION CountTable(table_name   IN VARCHAR2, 
                      where_clause IN VARCHAR2 DEFAULT NULL,
                      owner_id     IN VARCHAR2 DEFAULT USER) RETURN NUMBER IS
  cnt           NUMBER;
  source_cursor INTEGER;
  ignore        INTEGER;
  BEGIN
    Debug('Start CountTable');
    source_cursor := dbms_sql.open_cursor;
    IF where_clause IS NULL THEN
      dbms_sql.parse(source_cursor,'SELECT count(*) FROM "' ||owner_id||'".'||table_name,dbms_sql.native);
    ELSE
      IF UPPER(SUBSTR(where_clause,1,5))!='WHERE' THEN
        dbms_sql.parse(source_cursor,'SELECT count(*) FROM "' || owner_id||'".'||table_name ||' WHERE '||where_clause,dbms_sql.native);
      ELSE
        dbms_sql.parse(source_cursor,'SELECT count(*) FROM "' || owner_id||'".'||table_name ||' '||where_clause,dbms_sql.native);
      END IF;
    END IF;
    dbms_sql.define_column(source_cursor, 1, cnt);
    ignore := dbms_sql.execute(source_cursor);
    IF dbms_sql.fetch_rows(source_cursor)>0 THEN
      dbms_sql.column_value(source_cursor, 1, cnt);
    END IF;
    dbms_sql.close_cursor(source_cursor);
    Debug('End CountTable');
    return cnt;
  EXCEPTION WHEN OTHERS THEN
    Debug('CountTable'||sqlerrm);
    IF dbms_sql.is_open(source_cursor) THEN
      dbms_sql.close_cursor(source_cursor);
    END IF;
    RAISE;
    return 0;
  END;  

  PROCEDURE GetTableDefinition(tn       IN  VARCHAR2,  
                               tdef     OUT TableDef,
                               tcol     OUT TableColumns,
                               tind     OUT TableIndexes,
                               tindcol  OUT TableIndexColumns,
                               tcon     OUT TableConstraints,
                               tconcol  OUT TableConstraintColumns,
                               ttrig    OUT TableTriggers,
                               ttrigcol OUT TableTriggerColumns,
                               tcolcom  OUT TableColumnComments,
                               tcom     OUT TableComments,
                               tfcon    OUT TableFConstraints,
                               tfconcol OUT TableFConstraintColumns,
                               tsyn     OUT TableSynonyms,
                               tgrnt    OUT TableGrants,
                               tcgrnt   OUT TableColumnGrants,
                               owner_id IN  VARCHAR2 DEFAULT USER) IS
  BEGIN
    Debug('Start GetTableDefinition');
    GetTableDef(tn,tdef,owner_id);
    GetTableColumns(tn,tcol,owner_id);
    GetTableIndexes(tn,tind,owner_id);
    GetTableIndexColumns(tn,tindcol,owner_id);
    GetTableConstraints(tn,tcon,owner_id);
    GetTableConstraintColumns(tn,tconcol,owner_id);
    GetTableTriggers(tn,ttrig,owner_id);
    GetTableColumnComments(tn,tcolcom,owner_id);
    GetTableComments(tn,tcom,owner_id);
    GetTableFConstraints(tn,tfcon,owner_id);
    GetTableFConstraintColumns(tn,tfconcol,owner_id);
    GetTableSynonyms(tn,tsyn,owner_id);
    GetTableGrants(tn,tgrnt,owner_id);
    GetTableColumnGrants(tn,tcgrnt,owner_id);
    Debug('End GetTableDefinition');
  END;
  
  PROCEDURE DropTable(table_name IN VARCHAR2,
                      owner_id   IN VARCHAR2 DEFAULT USER) IS
  BEGIN
    ClearLog;
    Debug('Start DropTable');
    WorkingLog('Dropping table '||upper(table_name));
    DDL('drop table "'||owner_id||'".'||table_name||' cascade constraints');
    Debug('End DropTable');
    PrintLog;
  END;

  PROCEDURE DropColumn(tn       IN VARCHAR2, 
                       cn       IN VARCHAR2,
                       owner_id IN VARCHAR2 DEFAULT USER) IS
  tdef     TableDef;
  tcol     TableColumns;
  tind     TableIndexes;
  tindcol  TableIndexColumns;
  tcon     TableConstraints;
  tconcol  TableConstraintColumns;
  ttrig    TableTriggers;
  ttrigcol TableTriggerColumns;
  tcolcom  TableColumnComments;
  tcom     TableComments;
  tfcon    TableFConstraints;
  tfconcol TableFConstraintColumns;
  tsyn     TableSynonyms;
  tgrnt    TableGrants;
  tcgrnt   TableColumnGrants;
  x        pls_integer;
  BEGIN
    ClearLog;
    Debug('Start DropColumn');
    GetTableDefinition(tn,tdef,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol,tsyn,tgrnt,tcgrnt,owner_id);
    if tcol.count=1 then -- Only one column in table
      ApplicationError(-20001,'There must be at least one column left in a table');
      PrintLog;
      return;
    end if;
    if tcol.count=0 then -- Table does not exist
      ApplicationError(-20002,'Table '||upper(tn)||' does not exist');
      PrintLog;
      return;
    end if;
    select count(*) into x from sys.all_tab_columns where table_name=upper(tn) and column_name=upper(cn) and owner=upper(owner_id);
    if x=0 then
      ApplicationError(-20005,'Column '||upper(tn)||'.'||upper(cn)||' does not exist');
      PrintLog;
      return;
    end if;
    RemoveColumn(cn,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol);
    RecreateTable(tn,cn,cn,null,null,tdef,tcol,tind,tindcol,tcon,tconcol,ttrig,ttrigcol,tcolcom,tcom,tfcon,tfconcol,tsyn,tgrnt,tcgrnt);
    Debug('End DropColumn');
    PrintLog;
  END;

  PROCEDURE DropConstraint(constraint_name IN VARCHAR2,
                           owner_id        IN VARCHAR2 DEFAULT USER) IS
  tc TableConstraints;
  tcc TableConstraintColumns;
  BEGIN
    ClearLog;
    Debug('Start DropConstraint');
    if not DropConstraint_I(constraint_name,owner_id,FALSE,tc,tcc) then
      PrintLog;
      return;
    end if;
    Debug('End DropConstraint');
    PrintLog;
  END;

  PROCEDURE DropIndex(index_name IN VARCHAR2,
                      owner_id   IN VARCHAR2 DEFAULT USER) IS
  tc TableConstraints;
  tcc TableConstraintColumns;
  BEGIN
    ClearLog;
    Debug('Start DropIndex');
    if not DropIndex_I(index_name,FALSE,tc,tcc,owner_id) then
      PrintLog;
      return;
    end if;
    Debug('End DropIndex');
    PrintLog;
  END;

  PROCEDURE DropTrigger(trigger_name IN VARCHAR2,
                        owner_id     IN VARCHAR2 DEFAULT USER) IS
  BEGIN
    ClearLog;
    Debug('Start DropTrigger');
    if not DropTrigger_I(trigger_name,owner_id) then
      PrintLog;
      return;
    end if;
    Debug('End DropTrigger');
    PrintLog;
  END;

  PROCEDURE DropColumnComment(table_name  IN VARCHAR2,
                              column_name IN VARCHAR2,
                              owner_id    IN VARCHAR2 DEFAULT USER) IS
  BEGIN
    ClearLog;
    Debug('Start DropColumnComment');
    if not DropColumnComment_I(table_name,column_name,owner_id) then
      PrintLog;
      return;
    end if;
    Debug('End DropColumnComment');
    PrintLog;
  END;

  PROCEDURE DropTableComment(table_name IN VARCHAR2,
                             owner_id   IN VARCHAR2 DEFAULT USER) IS
  BEGIN
    ClearLog;
    Debug('Start DropTableComment');
    if not DropTableComment_I(table_name,owner_id) then
      PrintLog;
      return;
    end if;
    Debug('End DropTableComment');
    PrintLog;
  END;

  PROCEDURE AlterTable(tn              IN VARCHAR2,
  		       table_name      IN VARCHAR2 DEFAULT NULL,
  		       tablespace_name IN VARCHAR2 DEFAULT NULL,
  		       pct_free        IN NUMBER   DEFAULT NULL,
  		       pct_used        IN NUMBER   DEFAULT NULL,
  		       ini_trans       IN NUMBER   DEFAULT NULL,
  		       max_trans       IN NUMBER   DEFAULT NULL,
  		       initial_extent  IN NUMBER   DEFAULT NULL,
  		       next_extent     IN NUMBER   DEFAULT NULL,
  		       min_extents     IN NUMBER   DEFAULT NULL,
  		       max_extents     IN NUMBER   DEFAULT NULL,
  		       pct_increase    IN NUMBER   DEFAULT NULL,
  		       freelists       IN NUMBER   DEFAULT NULL,
  		       freelist_groups IN NUMBER   DEFAULT NULL,
  		       owner_id        IN VARCHAR2 DEFAULT USER,
                       new_owner       IN VARCHAR2 DEFAULT NULL) IS
  n TableDef;
  o TableDef;
  cnt NUMBER;
  BEGIN
    ClearLog;
    Debug('Start AlterTable');
    select count(1) into cnt from all_tables where table_name=upper(tn) and owner=upper(owner_id);
    if cnt=0 then -- Table does not exist
      ApplicationError(-20002,'Table '||upper(tn)||' does not exist');
      PrintLog;
      return;
    end if;
    GetTableDef(tn,n,owner_id);
    o:=n;
    n(1).table_name:=nvl(upper(table_name),n(1).table_name);
    n(1).owner:=nvl(upper(new_owner),n(1).owner);
    n(1).tablespace_name:=nvl(upper(tablespace_name),n(1).tablespace_name);
    n(1).pct_free:=nvl(pct_free,n(1).pct_free);
    n(1).pct_used:=nvl(pct_used,n(1).pct_used);
    n(1).ini_trans:=nvl(ini_trans,n(1).ini_trans);
    n(1).max_trans:=nvl(max_trans,n(1).max_trans);
    n(1).initial_extent:=nvl(initial_extent,n(1).initial_extent);
    n(1).next_extent:=nvl(next_extent,n(1).next_extent);
    n(1).min_extents:=nvl(min_extents,n(1).min_extents);
    n(1).max_extents:=nvl(max_extents,n(1).max_extents);
    n(1).pct_increase:=nvl(pct_increase,n(1).pct_increase);
    n(1).freelists:=nvl(freelists,n(1).freelists);
    n(1).freelist_groups:=nvl(freelist_groups,n(1).freelist_groups);
    if not AlterTable_I(tn,o,n) then
      PrintLog;
      return;
    end if;
    Debug('End AlterTable');
    PrintLog;
  END;

  PROCEDURE AlterTables(tn                IN VARCHAR2 DEFAULT NULL,
                        s_table_name      IN VARCHAR2 DEFAULT NULL,
                        s_tablespace_name IN VARCHAR2 DEFAULT NULL,
                        s_pct_free        IN NUMBER   DEFAULT NULL,
                        s_pct_used        IN NUMBER   DEFAULT NULL,
                        s_ini_trans       IN NUMBER   DEFAULT NULL,
                        s_max_trans       IN NUMBER   DEFAULT NULL,
                        s_initial_extent  IN NUMBER   DEFAULT NULL,
                        s_next_extent     IN NUMBER   DEFAULT NULL,
                        s_min_extents     IN NUMBER   DEFAULT NULL,
                        s_max_extents     IN NUMBER   DEFAULT NULL,
                        s_pct_increase    IN NUMBER   DEFAULT NULL,
                        s_freelists       IN NUMBER   DEFAULT NULL,
                        s_freelist_groups IN NUMBER   DEFAULT NULL,
                        tablespace_name   IN VARCHAR2 DEFAULT NULL,
                        pct_free          IN NUMBER   DEFAULT NULL,
                        pct_used          IN NUMBER   DEFAULT NULL,
                        ini_trans         IN NUMBER   DEFAULT NULL,
                        max_trans         IN NUMBER   DEFAULT NULL,
                        initial_extent    IN NUMBER   DEFAULT NULL,
                        next_extent       IN NUMBER   DEFAULT NULL,
                        min_extents       IN NUMBER   DEFAULT NULL,
                        max_extents       IN NUMBER   DEFAULT NULL,
                        pct_increase      IN NUMBER   DEFAULT NULL,
                        freelists         IN NUMBER   DEFAULT NULL,
                        freelist_groups   IN NUMBER   DEFAULT NULL,
                        s_owner_id        IN VARCHAR2 DEFAULT USER,
                        owner_id          IN VARCHAR2 DEFAULT NULL) IS
  n TableDef;
  o TableDef;
  cursor c is select table_name,owner from all_tables
    where (s_table_name is null or table_name like upper(s_table_name))
    and (s_owner_id is null or owner like upper(s_owner_id))
    and (s_tablespace_name is null or tablespace_name like upper(s_tablespace_name))
    and (s_pct_free is null or pct_free=s_pct_free)
    and (s_pct_used is null or pct_used=s_pct_used)
    and (s_ini_trans is null or ini_trans=s_ini_trans)
    and (s_max_trans is null or max_trans=s_max_trans)
    and (s_initial_extent is null or initial_extent=s_initial_extent)
    and (s_next_extent is null or next_extent=s_next_extent)
    and (s_min_extents is null or min_extents=s_min_extents)
    and (s_max_extents is null or max_extents=s_max_extents)
    and (s_pct_increase is null or pct_increase=s_pct_increase)
    and (s_freelists is null or freelists=s_freelists)
    and (s_freelist_groups is null or freelist_groups=s_freelist_groups)
    order by table_name;
  BEGIN
    ClearLog;
    Debug('Start AlterTables');
    for x in c loop
      WorkingLog('Altering table '||x.table_name);
      GetTableDef(x.table_name,n,x.owner);
      o:=n;
      n(1).tablespace_name:=nvl(upper(tablespace_name),n(1).tablespace_name);
      n(1).pct_free:=nvl(pct_free,n(1).pct_free);
      n(1).owner:=nvl(owner_id,n(1).owner);
      n(1).pct_used:=nvl(pct_used,n(1).pct_used);
      n(1).ini_trans:=nvl(ini_trans,n(1).ini_trans);
      n(1).max_trans:=nvl(max_trans,n(1).max_trans);
      n(1).initial_extent:=nvl(initial_extent,n(1).initial_extent);
      n(1).next_extent:=nvl(next_extent,n(1).next_extent);
      n(1).min_extents:=nvl(min_extents,n(1).min_extents);
      n(1).max_extents:=nvl(max_extents,n(1).max_extents);
      n(1).pct_increase:=nvl(pct_increase,n(1).pct_increase);
      n(1).freelists:=nvl(freelists,n(1).freelists);
      n(1).freelist_groups:=nvl(freelist_groups,n(1).freelist_groups);
      if not AlterTable_I(x.table_name,o,n) then
        PrintLog;
        return;
      end if;
    end loop;
    Debug('End AlterTables');
    PrintLog;
  END;

  PROCEDURE AlterIndex(ind             IN VARCHAR2,
  		       index_name      IN VARCHAR2 DEFAULT NULL,
  		       index_type      IN VARCHAR2 DEFAULT NULL,
 	               tablespace_name IN VARCHAR2 DEFAULT NULL,
 	               uniqueness      IN BOOLEAN  DEFAULT NULL,
  		       pct_free        IN NUMBER   DEFAULT NULL,
  		       ini_trans       IN NUMBER   DEFAULT NULL,
  		       max_trans       IN NUMBER   DEFAULT NULL,
  		       initial_extent  IN NUMBER   DEFAULT NULL,
  		       next_extent     IN NUMBER   DEFAULT NULL,
  		       min_extents     IN NUMBER   DEFAULT NULL,
  		       max_extents     IN NUMBER   DEFAULT NULL,
  		       pct_increase    IN NUMBER   DEFAULT NULL,
  		       freelists       IN NUMBER   DEFAULT NULL,
  		       freelist_groups IN NUMBER   DEFAULT NULL,
  		       index_columns   IN TableIndexColumns,
                       owner_id        IN VARCHAR2 DEFAULT USER,
                       new_owner       IN VARCHAR2 DEFAULT NULL) IS
  n TableIndexes;
  o TableIndexes;
  ic TableIndexColumns;
  cnt NUMBER;
  BEGIN
    ClearLog;
    Debug('Start AlterIndex');
    select count(1) into cnt from all_indexes where index_name=upper(ind) and owner=upper(owner_id);
    if cnt=0 then -- Index does not exist
      ApplicationError(-20002,'Index '||upper(ind)||' does not exist');
      PrintLog;
      return;
    end if;
    GetIndexDef(ind,n,owner_id);
    GetIndexColumns(ind,ic,owner_id);
    o:=n;
    n(1).index_name:=nvl(upper(index_name),n(1).index_name);
    n(1).owner:=nvl(upper(new_owner),n(1).owner);
    n(1).index_type:=nvl(upper(index_type),n(1).index_type);
    n(1).tablespace_name:=nvl(upper(tablespace_name),n(1).tablespace_name);
    n(1).pct_free:=nvl(pct_free,n(1).pct_free);
    n(1).ini_trans:=nvl(ini_trans,n(1).ini_trans);
    n(1).max_trans:=nvl(max_trans,n(1).max_trans);
    n(1).initial_extent:=nvl(initial_extent,n(1).initial_extent);
    n(1).next_extent:=nvl(next_extent,n(1).next_extent);
    n(1).min_extents:=nvl(min_extents,n(1).min_extents);
    n(1).max_extents:=nvl(max_extents,n(1).max_extents);
    n(1).pct_increase:=nvl(pct_increase,n(1).pct_increase);
    n(1).freelists:=nvl(freelists,n(1).freelists);
    n(1).freelist_groups:=nvl(freelist_groups,n(1).freelist_groups);
    if uniqueness is not null then
      if uniqueness then
        n(1).uniqueness:='TRUE';
      else
        n(1).uniqueness:='FALSE';
      end if;
    end if;
    if not AlterIndex_I(ind,o,n,ic,index_columns) then
      select count(1) into cnt from all_indexes where index_name=upper(ind) and owner=owner_id;
      if cnt=0 then
        WorkingLog('Alter index failed. Recreating old index...');
        CreateIndexes(ind,o,ic);
      end if;
      PrintLog;
      return;
    end if;
    Debug('End AlterIndex');
    PrintLog;
  END;

  PROCEDURE AlterIndex(ind             IN VARCHAR2,
  		       index_name      IN VARCHAR2 DEFAULT NULL,
  		       index_type      IN VARCHAR2 DEFAULT NULL,
 	               tablespace_name IN VARCHAR2 DEFAULT NULL,
 	               uniqueness      IN BOOLEAN  DEFAULT NULL,
  		       pct_free        IN NUMBER   DEFAULT NULL,
  		       ini_trans       IN NUMBER   DEFAULT NULL,
  		       max_trans       IN NUMBER   DEFAULT NULL,
  		       initial_extent  IN NUMBER   DEFAULT NULL,
  		       next_extent     IN NUMBER   DEFAULT NULL,
  		       min_extents     IN NUMBER   DEFAULT NULL,
  		       max_extents     IN NUMBER   DEFAULT NULL,
  		       pct_increase    IN NUMBER   DEFAULT NULL,
  		       freelists       IN NUMBER   DEFAULT NULL,
  		       freelist_groups IN NUMBER   DEFAULT NULL,
                       owner_id        IN VARCHAR2 DEFAULT USER,
                       new_owner_id    IN VARCHAR2 DEFAULT NULL) IS
  n TableIndexes;
  o TableIndexes;
  ic TableIndexColumns;
  cnt NUMBER;
  BEGIN
    ClearLog;
    Debug('Start AlterIndex');
    select count(1) into cnt from all_indexes where index_name=upper(ind) and owner=upper(owner_id);
    if cnt=0 then -- Index does not exist
      ApplicationError(-20002,'Index '||upper(ind)||' does not exist');
      PrintLog;
      return;
    end if;
    GetIndexDef(ind,n,owner_id);
    GetIndexColumns(ind,ic,owner_id);
    o:=n;
    n(1).index_name:=nvl(upper(index_name),n(1).index_name);
    n(1).owner:=nvl(upper(new_owner_id),n(1).owner);
    n(1).index_type:=nvl(upper(index_type),n(1).index_type);
    n(1).tablespace_name:=nvl(upper(tablespace_name),n(1).tablespace_name);
    n(1).pct_free:=nvl(pct_free,n(1).pct_free);
    n(1).ini_trans:=nvl(ini_trans,n(1).ini_trans);
    n(1).max_trans:=nvl(max_trans,n(1).max_trans);
    n(1).initial_extent:=nvl(initial_extent,n(1).initial_extent);
    n(1).next_extent:=nvl(next_extent,n(1).next_extent);
    n(1).min_extents:=nvl(min_extents,n(1).min_extents);
    n(1).max_extents:=nvl(max_extents,n(1).max_extents);
    n(1).pct_increase:=nvl(pct_increase,n(1).pct_increase);
    n(1).freelists:=nvl(freelists,n(1).freelists);
    n(1).freelist_groups:=nvl(freelist_groups,n(1).freelist_groups);
    if uniqueness is not null then
      if uniqueness then
        n(1).uniqueness:='TRUE';
      else
        n(1).uniqueness:='FALSE';
      end if;
    end if;
    if not AlterIndex_I(ind,o,n,ic,ic) then
      select count(1) into cnt from all_indexes where index_name=upper(ind) and owner=owner_id;
      if cnt=0 then
        WorkingLog('Alter index failed. Recreating old index...');
        CreateIndexes(ind,o,ic);
      end if;
      PrintLog;
      return;
    end if;
    Debug('End AlterIndex');
    PrintLog;
  END;

  PROCEDURE AlterIndexes(s_index_name      IN VARCHAR2 DEFAULT NULL,
                         s_index_type      IN VARCHAR2 DEFAULT NULL,
                         s_tablespace_name IN VARCHAR2 DEFAULT NULL,
                         s_uniqueness      IN BOOLEAN  DEFAULT NULL,
                         s_pct_free        IN NUMBER   DEFAULT NULL,
                         s_ini_trans       IN NUMBER   DEFAULT NULL,
                         s_max_trans       IN NUMBER   DEFAULT NULL,
                         s_initial_extent  IN NUMBER   DEFAULT NULL,
                         s_next_extent     IN NUMBER   DEFAULT NULL,
                         s_min_extents     IN NUMBER   DEFAULT NULL,
                         s_max_extents     IN NUMBER   DEFAULT NULL,
                         s_pct_increase    IN NUMBER   DEFAULT NULL,
                         s_freelists       IN NUMBER   DEFAULT NULL,
                         s_freelist_groups IN NUMBER   DEFAULT NULL,
                         index_type        IN VARCHAR2 DEFAULT NULL,
                         tablespace_name   IN VARCHAR2 DEFAULT NULL,
                         uniqueness        IN BOOLEAN  DEFAULT NULL,
                         pct_free          IN NUMBER   DEFAULT NULL,
                         ini_trans         IN NUMBER   DEFAULT NULL,
                         max_trans         IN NUMBER   DEFAULT NULL,
                         initial_extent    IN NUMBER   DEFAULT NULL,
                         next_extent       IN NUMBER   DEFAULT NULL,
                         min_extents       IN NUMBER   DEFAULT NULL,
                         max_extents       IN NUMBER   DEFAULT NULL,
                         pct_increase      IN NUMBER   DEFAULT NULL,
                         freelists         IN NUMBER   DEFAULT NULL,
                         freelist_groups   IN NUMBER   DEFAULT NULL,
                         s_owner_id        IN VARCHAR2 DEFAULT USER,
                         owner_id          IN VARCHAR2 DEFAULT NULL) IS
  n TableIndexes;
  o TableIndexes;
  ic TableIndexColumns;
  cnt NUMBER;
  suq VARCHAR2(10);
  cursor c is select index_name,owner from all_indexes
    where (s_index_name is null or index_name like upper(s_index_name))
    and (s_tablespace_name is null or tablespace_name like upper(s_tablespace_name))
    and (s_owner_id is null or owner like upper(s_owner_id))
    and (s_pct_free is null or pct_free=s_pct_free)
    and (s_ini_trans is null or ini_trans=s_ini_trans)
    and (s_max_trans is null or max_trans=s_max_trans)
    and (s_initial_extent is null or initial_extent=s_initial_extent)
    and (s_next_extent is null or next_extent=s_next_extent)
    and (s_min_extents is null or min_extents=s_min_extents)
    and (s_max_extents is null or max_extents=s_max_extents)
    and (s_pct_increase is null or pct_increase=s_pct_increase)
    and (s_freelists is null or freelists=s_freelists)
    and (s_freelist_groups is null or freelist_groups=s_freelist_groups)
    and (s_index_type is null or index_type=upper(s_index_type))
    and (suq is null or uniqueness=suq)
    order by index_name;
  BEGIN
    ClearLog;
    Debug('Start AlterIndexes');
    if s_uniqueness is not null then
      if s_uniqueness then
        suq:='TRUE';
      else
        suq:='FALSE';
      end if;
    end if;
    for x in c loop
      n.delete;
      ic.delete;
      o.delete;
      GetIndexDef(x.index_name,n,x.owner);
      GetIndexColumns(x.index_name,ic,x.owner);
      o:=n;
      n(1).owner:=nvl(upper(owner_id),n(1).owner);
      n(1).index_type:=nvl(upper(index_type),n(1).index_type);
      n(1).tablespace_name:=nvl(upper(tablespace_name),n(1).tablespace_name);
      n(1).pct_free:=nvl(pct_free,n(1).pct_free);
      n(1).ini_trans:=nvl(ini_trans,n(1).ini_trans);
      n(1).max_trans:=nvl(max_trans,n(1).max_trans);
      n(1).initial_extent:=nvl(initial_extent,n(1).initial_extent);
      n(1).next_extent:=nvl(next_extent,n(1).next_extent);
      n(1).min_extents:=nvl(min_extents,n(1).min_extents);
      n(1).max_extents:=nvl(max_extents,n(1).max_extents);
      n(1).pct_increase:=nvl(pct_increase,n(1).pct_increase);
      n(1).freelists:=nvl(freelists,n(1).freelists);
      n(1).freelist_groups:=nvl(freelist_groups,n(1).freelist_groups);
      if uniqueness is not null then
        if uniqueness then
          n(1).uniqueness:='TRUE';
        else
          n(1).uniqueness:='FALSE';
        end if;
      end if;
      if not AlterIndex_I(x.index_name,o,n,ic,ic) then
        select count(1) into cnt from all_indexes where index_name=upper(x.index_name) and owner=upper(owner_id);
        if cnt=0 then
          select count(*) into cnt from all_indexes where index_name=o(1).index_name and owner=o(1).owner;
          if cnt=0 then
            WorkingLog('Alter index failed. Recreating old index...');
            CreateIndexes(x.index_name,o,ic);
          end if;
        end if;
      end if;
    end loop;
    Debug('End AlterIndexes');
    PrintLog;
  END;

  PROCEDURE AlterConstraint(cn                IN VARCHAR2,
       		            constraint_name   IN VARCHAR2 DEFAULT NULL,
                            constraint_type   IN VARCHAR2 DEFAULT NULL, -- P, C, R, U
                            table_name        IN VARCHAR2 DEFAULT NULL,
                            search_condition  IN VARCHAR2 DEFAULT NULL,
                            r_owner           IN VARCHAR2 DEFAULT NULL,
                            r_constraint_name IN VARCHAR2 DEFAULT NULL,
                            delete_rule       IN VARCHAR2 DEFAULT NULL, -- CASCADE, NO_ACTION
                            status            IN VARCHAR2 DEFAULT NULL, -- ENABLED, DISABLED
                            deferrable        IN VARCHAR2 DEFAULT NULL, -- DEFERRABLE, NOT DEFERRABLE
                            deferred          IN VARCHAR2 DEFAULT NULL, -- IMMEDIATE, DEFERRED
                            validated         IN VARCHAR2 DEFAULT NULL, -- VALIDATED, NOT VALIDATED
                            owner_id          IN VARCHAR2 DEFAULT USER,
                            new_owner_id      IN VARCHAR2 DEFAULT NULL) IS
  cnt NUMBER;
  o   TableConstraints;
  n   TableConstraints;
  tc  TableConstraints;
  tcc TableConstraintColumns;
  cc  TableConstraintColumns;
  BEGIN
    ClearLog;
    Debug('Start AlterConstraint');
    select count(1) into cnt from all_constraints where constraint_name=upper(cn) and owner=upper(owner_id);
    if cnt=0 then -- Index does not exist
      ApplicationError(-20002,'Constraint '||upper(cn)||' does not exist');
      PrintLog;
      return;
    end if;
    GetConstraintDef(cn,o,owner_id);
    GetConstraintColumns(cn,cc,owner_id);
    n:=o;
    n(1).constraint_name:=nvl(upper(constraint_name),o(1).constraint_name);
    n(1).constraint_type:=nvl(upper(constraint_type),o(1).constraint_type);
    n(1).table_name:=nvl(upper(table_name),o(1).table_name);
    n(1).search_condition:=nvl(upper(search_condition),o(1).search_condition);
    n(1).r_owner:=nvl(upper(r_owner),o(1).r_owner);
    n(1).r_constraint_name:=nvl(upper(r_constraint_name),o(1).r_constraint_name);
    n(1).delete_rule:=nvl(upper(delete_rule),o(1).delete_rule);
    n(1).status:=nvl(upper(status),o(1).status);
    n(1).deferrable:=nvl(upper(deferrable),o(1).deferrable);
    n(1).deferred:=nvl(upper(deferred),o(1).deferred);
    n(1).validated:=nvl(upper(validated),o(1).validated);
    n(1).owner:=nvl(upper(new_owner_id),o(1).owner);
    if n(1).constraint_name!=o(1).constraint_name 
     or n(1).constraint_type!=o(1).constraint_type
     or n(1).table_name!=o(1).table_name
     or n(1).search_condition!=o(1).search_condition
     or n(1).r_owner!=o(1).r_owner
     or n(1).r_constraint_name!=o(1).r_constraint_name
     or n(1).delete_rule!=o(1).delete_rule
     or n(1).deferrable!=o(1).deferrable
     or n(1).deferred!=o(1).deferred
     or n(1).validated!=o(1).validated
     or n(1).owner!=o(1).owner then
      if not DropConstraint_I(cn,owner_id,FALSE,tc,tcc) then
        PrintLog;
        return;
      end if;
      CreateConstraints(n,cc);
    elsif n(1).status!=o(1).status then
      if n(1).status='ENABLED' then
        DDL('alter table "'||n(1).owner||'".'||n(1).table_name||' enable constraint '||n(1).constraint_name);
      elsif n(1).status='DISABLED' then
        DDL('alter table "'||n(1).owner||'".'||n(1).table_name||' disable constraint '||n(1).constraint_name);
      end if;
    end if;
    Debug('End AlterConstraint');
    PrintLog;
  END;

  PROCEDURE AlterConstraint(cn                IN VARCHAR2,
                            cons_columns      IN TableConstraintColumns,
       		            constraint_name   IN VARCHAR2 DEFAULT NULL,
                            constraint_type   IN VARCHAR2 DEFAULT NULL, -- P, C, R, U
                            table_name        IN VARCHAR2 DEFAULT NULL,
                            search_condition  IN VARCHAR2 DEFAULT NULL,
                            r_owner           IN VARCHAR2 DEFAULT NULL,
                            r_constraint_name IN VARCHAR2 DEFAULT NULL,
                            delete_rule       IN VARCHAR2 DEFAULT NULL, -- CASCADE, NO_ACTION
                            status            IN VARCHAR2 DEFAULT NULL, -- ENABLED, DISABLED
                            deferrable        IN VARCHAR2 DEFAULT NULL, -- DEFERRABLE, NOT DEFERRABLE
                            deferred          IN VARCHAR2 DEFAULT NULL, -- IMMEDIATE, DEFERRED
                            validated         IN VARCHAR2 DEFAULT NULL, -- VALIDATED, NOT VALIDATED
                            owner_id          IN VARCHAR2 DEFAULT USER,
                            new_owner_id      IN VARCHAR2 DEFAULT NULL) IS
  cnt NUMBER;
  o   TableConstraints;
  n   TableConstraints;
  tc  TableConstraints;
  tcc TableConstraintColumns;
  cc  TableConstraintColumns;
  same BOOLEAN:=TRUE;
  BEGIN
    ClearLog;
    Debug('Start AlterConstraint');
    select count(1) into cnt from all_constraints where constraint_name=upper(cn) and owner=upper(owner_id);
    if cnt=0 then -- Constraint does not exist
      ApplicationError(-20002,'Constraint '||upper(cn)||' does not exist');
      PrintLog;
      return;
    end if;
    GetConstraintDef(cn,o,owner_id);
    GetConstraintColumns(cn,cc,owner_id);
    n:=o;
    n(1).constraint_name:=nvl(upper(constraint_name),o(1).constraint_name);
    n(1).constraint_type:=nvl(upper(constraint_type),o(1).constraint_type);
    n(1).table_name:=nvl(upper(table_name),o(1).table_name);
    n(1).search_condition:=nvl(upper(search_condition),o(1).search_condition);
    n(1).r_owner:=nvl(upper(r_owner),o(1).r_owner);
    n(1).r_constraint_name:=nvl(upper(r_constraint_name),o(1).r_constraint_name);
    n(1).delete_rule:=nvl(upper(delete_rule),o(1).delete_rule);
    n(1).status:=nvl(upper(status),o(1).status);
    n(1).deferrable:=nvl(upper(deferrable),o(1).deferrable);
    n(1).deferred:=nvl(upper(deferred),o(1).deferred);
    n(1).validated:=nvl(upper(validated),o(1).validated);
    n(1).owner:=nvl(upper(new_owner_id),o(1).owner);
    for x in cc.first..cc.last loop
      if cc.exists(x) then
        if not cons_columns.exists(x) then
          same:=false;
          exit;
        end if;
        if cc(x).column_name!=cons_columns(x).column_name then
          same:=false;
          exit;
        end if;
      elsif cons_columns.exists(x) then
        same:=false;
        exit;
      end if;
    end loop;
    if n(1).constraint_name!=o(1).constraint_name 
     or n(1).constraint_type!=o(1).constraint_type
     or n(1).table_name!=o(1).table_name
     or n(1).search_condition!=o(1).search_condition
     or n(1).r_owner!=o(1).r_owner
     or n(1).r_constraint_name!=o(1).r_constraint_name
     or n(1).delete_rule!=o(1).delete_rule
     or n(1).deferrable!=o(1).deferrable
     or n(1).deferred!=o(1).deferred
     or n(1).validated!=o(1).validated
     or n(1).owner!=o(1).owner 
     or not same then
      if not DropConstraint_I(cn,owner_id,FALSE,tc,tcc) then
        PrintLog;
        return;
      end if;
      CreateConstraints(n,cons_columns);
    elsif n(1).status!=o(1).status then
      if n(1).status='ENABLED' then
        DDL('alter table "'||n(1).owner||'".'||n(1).table_name||' enable constraint '||n(1).constraint_name);
      elsif n(1).status='DISABLED' then
        DDL('alter table "'||n(1).owner||'".'||n(1).table_name||' disable constraint '||n(1).constraint_name);
      end if;
    end if;
    Debug('End AlterConstraint');
    PrintLog;
  END;
  
  PROCEDURE Grep(what     IN VARCHAR2,               
                 tn       IN VARCHAR2 DEFAULT NULL,  
                 cn       IN VARCHAR2 DEFAULT NULL,  
                 owner_id IN VARCHAR2 DEFAULT USER, 
		 casesens IN BOOLEAN  DEFAULT FALSE,
                 result   IN VARCHAR2 DEFAULT 'TABMAN_GREP_RESULTS') IS
    cursor c is select ata.table_name,atc.column_name,atc.data_type,ata.owner 
                from all_tab_columns atc,all_tables ata
                where (tn is null or ata.table_name like upper(tn))
                and (cn is null or atc.column_name like upper(cn))
                and (owner_id is null or ata.owner like upper(owner_id))
                and atc.data_type in ('VARCHAR2','CHAR','DATE','NUMBER') 
                and ata.table_name!=result
                and ata.table_name=atc.table_name
                and ata.owner=atc.owner
                order by ata.owner,ata.table_name,atc.column_name;
    cnt number;    
    ltn varchar2(30):='-1';
  BEGIN
    ClearLog;
    Debug('Start Grep');
    select count(*) into cnt from user_tables where table_name=upper(result);
    WorkingLog('Creating result table '||result);
    if cnt=0 then
      DDL('create table '||result||' (owner VARCHAR2(30),table_name VARCHAR2(30),column_name VARCHAR2(30),row_id ROWID,text VARCHAR2(4000))');
    end if;
    DDL('truncate table '||result);
    for x in c loop
      if ltn!=x.table_name then
        WorkingLog('Searching in table '||x.table_name);
        ltn:=x.table_name;
      end if;
      Grep_I(x.table_name,x.owner,x.column_name,what,casesens,result,x.data_type);
    end loop;
    PrintLog;
  END;
    

END;
/

show err
